# # type: ignore
# from typing import Dict, Union

# from openghg.store import ObsSurface
# from openghg.store.base import Datasource
# from openghg.util import create_daterange_str, verify_site


# class RankSources:
#     def __init__(self, cloud: bool = False) -> None:
#         self._cloud = cloud

#         raise NotImplementedError("Ranking currently not supported.")

#         if cloud:
#             raise NotImplementedError

#     def raw(self) -> Dict:
#         """Return the raw ranking data

#         Args:
#             None
#         Returns:
#             dict: Raw ranking data
#         """
#         return self._user_info

#     def get_sources(self, site: str, species: str) -> Dict:
#         """Get the datasources for this site and species to allow a ranking to be set

#         Args:
#             site: Three letter site code
#             species: Species name
#         Returns:
#             dict: Dictionary of datasource metadata
#         """
#         if self._cloud:
#             raise NotImplementedError
#             # return self._get_sources_cloud(site=site, species=species)
#         else:
#             return self._get_sources_local(site=site, species=species)

#     # def _get_sources_cloud(self, site: str, species: str) -> Dict:
#     #     site = verify_site(site=site)

#     #     args = {"site": site, "species": species}

#     #     self.site = site
#     #     self.species = species

#     #     response: Dict = self._service.call_function(function="rank.get_sources", args=args)

#     #     if not response:
#     #         raise ValueError(f"No sources found for {species} at {site}")

#     #     self._user_info: Dict = response["user_info"]
#     #     self._key_lookup: Dict = response["key_lookup"]
#     #     self._needs_update = False

#     #     return self._user_info

#     def _get_sources_local(self, site: str, species: str) -> Dict:
#         raise NotImplementedError
#         site = verify_site(site=site)

#         # Save these
#         self.site = site
#         self.species = species

#         obs = ObsSurface.load()
#         datasource_uuids = obs.datasources()
#         rank_table = obs.rank_data()

#         # Shallow load the Datasources (only get their JSON metadata)
#         datasources = (Datasource.load(uuid=uuid, shallow=True) for uuid in datasource_uuids)

#         matching_sources = [d for d in datasources if d.search_metadata(site=site, species=species)]

#         if matching_sources:
#             self._user_info = {
#                 d.inlet(): {
#                     "rank_data": rank_table.get(d.uuid(), "NA"),
#                     "data_range": d.daterange_str(),
#                 }
#                 for d in matching_sources
#             }

#             self._key_lookup = {d.inlet(): d.uuid() for d in matching_sources}
#             self._needs_update = False
#         else:
#             self._user_info = {}

#         return self._user_info

#     def get_specific_source(self, inlet: str) -> Dict:
#         """Return the ranking data of a specific key

#         Args:
#             key: Key
#         Returns:
#             dict: Dictionary of ranking data
#         """
#         if self._needs_update:
#             self.get_sources(site=self.site, species=self.species)

#         rank_data: Dict[str, Union[str, Dict]] = self._user_info[inlet]["rank_data"]
#         return rank_data

#     def set_rank(
#         self,
#         inlet: str,
#         rank: Union[int, str],
#         start_date: str,
#         end_date: str,
#         overwrite: bool = False,
#     ) -> None:
#         """Set the rank data for the

#         Args:
#             inlet: Inlet to set ranking data
#             rank: Number between 1 and 9
#             start_date: Start date
#             end_date: End date
#             overwrite: If True overwrite current ranking data
#         Returns:
#             None
#         """
#         if self._cloud:
#             raise NotImplementedError
#         else:
#             return self._set_rank_local(
#                 inlet=inlet,
#                 rank=rank,
#                 start_date=start_date,
#                 end_date=end_date,
#                 overwrite=overwrite,
#             )

#     # def _set_rank_cloud(
#     #     self,
#     #     inlet: str,
#     #     rank: Union[int, str],
#     #     start_date: str,
#     #     end_date: str,
#     #     overwrite: bool = False,
#     # ) -> None:
#     #     inlet = inlet.lower()
#     #     uuid = self._key_lookup[inlet]

#     #     dateranges = create_daterange_str(start=start_date, end=end_date)

#     #     args: Dict[str, Union[str, int, List]] = {}
#     #     args["rank"] = rank
#     #     args["uuid"] = uuid
#     #     args["dateranges"] = dateranges
#     #     args["overwrite"] = overwrite

#     #     self._service.call_function(function="rank.set_rank", args=args)
#     #     self._needs_update = True

#     def _set_rank_local(
#         self,
#         inlet: str,
#         rank: Union[int, str],
#         start_date: str,
#         end_date: str,
#         overwrite: bool = False,
#     ) -> None:
#         obs = ObsSurface.load()

#         inlet = inlet.lower()
#         uuid = self._key_lookup[inlet]

#         daterange = create_daterange_str(start=start_date, end=end_date)

#         obs.set_rank(uuid=uuid, rank=rank, date_range=daterange, overwrite=overwrite)

#         self._needs_update = True

#     def clear_rank(self, inlet: str) -> None:
#         """Clear the ranking data for a Datasource

#         Args:
#             key: Key for specific source
#         Returns:
#             None
#         """
#         if self._cloud:
#             raise NotImplementedError
#             # return self._clear_rank_cloud(inlet=inlet)
#         else:
#             return self._clear_rank_local(inlet=inlet)

#     def _clear_rank_local(self, inlet: str) -> None:
#         """Clear the ranking data for a Datasource

#         Args:
#             key: Key for specific source
#         Returns:
#             None
#         """
#         obs = ObsSurface.load()
#         inlet = inlet.lower()
#         uuid = self._key_lookup[inlet]
#         obs.clear_rank(uuid=uuid)
#         self._needs_update = True

#     # def _clear_rank_cloud(self, inlet: str) -> None:
#     #     """Clear the ranking data for a Datasource

#     #     Args:
#     #         key: Key for specific source
#     #     Returns:
#     #         None
#     #     """
#     #     uuid = self._key_lookup[inlet]
#     #     args = {"uuid": uuid}
#     #     self._service.call_function(function="rank.clear_rank", args=args)
#     #     self._needs_update = True

#     # def visualise_rankings(self) -> Network:
#     #     """ Creates a small network graph of ranked data with each rank given a colour

#     #         Note that this function should only be run from a Jupyter Notebook

#     #         Args:
#     #             rank_data (dict): Dictionary of the form given by RankSources.get_sources()
#     #         Returns:
#     #             pyvis.network.Network: Network graph
#     #     """
#     #     header_text = "OpenGHG ranked data"
#     #     net = Network("800px", "100%", notebook=True, heading=header_text)
#     #     # Set the physics layout of the network
#     #     net.force_atlas_2based()

#     #     rank_data = self._key_lookup

#     #     a_key = list(rank_data.keys())[0]
#     #     site = rank_data[a_key]["metadata"]["site"].upper()

#     #     norm = matplotlib.colors.Normalize(vmin=0, vmax=10, clip=True)
#     #     mapper = cm.ScalarMappable(norm=norm, cmap=cm.tab10)

#     #     def colour_mapper(x):
#     #         return matplotlib.colors.to_hex(mapper.to_rgba(int(x)))

#     #     net.add_node(site, label=site, color="brown", value=5000)

#     #     for key, data in rank_data.items():
#     #         rank = data["rank"]
#     #         site_name = data["metadata"]["site"].upper()
#     #         data_range = data["data_range"]

#     #         # HTML to show when the mouse is hovered over a node
#     #         title = "</br>".join([f"<b>Rank:</b> {str(rank)}", f"<b>Site:</b> {site_name}", f"<b>Data range:</b> {data_range}"])

#     #         if rank == 0:
#     #             colour = colour_mapper(rank)
#     #         else:
#     #             # For now just use the highest rank for the color
#     #             highest_rank = sorted(list(rank.keys()))[-1]
#     #             colour = colour_mapper(highest_rank)

#     #         split_key = key.split("_")
#     #         label = " ".join((split_key[0].upper(), split_key[1].upper(), split_key[2], split_key[3]))

#     #         net.add_node(key, label=label, title=title, color=colour, value=2000)
#     #         net.add_edge(source=site, to=key)

#     #     return net.show("openghg_rankings.html")



import logging
import math
from copy import deepcopy
from typing import Dict, List, Tuple, Optional
from openghg.types import AttrMismatchError
from openghg.util import is_number

logger = logging.getLogger("openghg.standardise.metadata")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


def metadata_default_keys() -> List:
    """
    Defines default values expected within ObsSurface metadata.
    Returns:
        list: keys required in metadata
    """
    default_keys = [
        "site",
        "species",
        "inlet",
        "inlet_height_magl",
        "network",
        "instrument",
        "sampling_period",
        "calibration_scale",
        "data_owner",
        "data_owner_email",
        "station_longitude",
        "station_latitude",
        "station_long_name",
        "station_height_masl",
    ]

    return default_keys


def metadata_keys_as_floats() -> List:
    """
    Defines which keys should be consistently stored as numbers in the metadata
    (even if they are not numbers within the attributes).
    Returns:
        list: keys required to be floats in metadata
    """

    values_as_floats = [
        # "inlet_height_magl",
        "station_longitude",
        "station_latitude",
        "station_height_masl",
    ]

    return values_as_floats


def sync_surface_metadata(
    metadata: Dict,
    attributes: Dict,
    keys_to_add: Optional[List] = None,
    update_mismatch: str = "never",
) -> Tuple[Dict, Dict]:
    """
    Makes sure any duplicated keys between the metadata and attributes
    dictionaries match and that certain keys are present in the metadata.

    Args:
        metadata: Dictionary of metadata
        attributes: Attributes
        keys_to_add: Add these keys to the metadata, if not present, based on
        the attribute values. Note: this skips any keys which can't be
        copied from the attribute values.
        update_mismatch: If case insensitive mismatch is found between an
          attribute and a metadata value, this determines the function behaviour.
          This includes the options:
            - "never" - don't update mismatches and raise an AttrMismatchError
            - "from_source" / "attributes" - update mismatches based on input attributes
            - "from_definition" / "metadata" - update mismatches based on input metadata
    Returns:
        dict, dict: Aligned metadata, attributes
    """
    meta_copy = deepcopy(metadata)
    attrs_copy = deepcopy(attributes)

    mismatch_keys = {
        "never": ["never"],
        "attributes": ["attributes", "from_source"],
        "metadata": ["metadata", "from_definition"],
    }

    for key, options in mismatch_keys.items():
        if update_mismatch.lower() in options:
            update_mismatch = key.lower()
            break
    else:
        raise ValueError(f"Input for 'update_mismatch' should be one of {mismatch_keys}")

    attr_mismatches = {}

    # Check if we have differences
    for key, meta_value in metadata.items():
        try:
            attr_value = attributes[key]

            # This should mainly be used for lat/long
            relative_tolerance = 1e-3

            if is_number(attr_value) and is_number(meta_value):
                if not math.isclose(float(attr_value), float(meta_value), rel_tol=relative_tolerance):
                    err_warn_num = f"Value of {key} not within tolerance, metadata: {meta_value} - attributes: {attr_value}"
                    if update_mismatch == "never":
                        attr_mismatches[key] = (meta_value, attr_value)
                    elif update_mismatch == "attributes":
                        logger.warning(
                            f"{err_warn_num}\nUpdating metadata to use attribute value of {key} = {attr_value}"
                        )
                        meta_copy[key] = str(attr_value)
                    elif update_mismatch == "metadata":
                        logger.warning(
                            f"{err_warn_num}\nUpdating attributes to use metadata value of {key} = {meta_value}"
                        )
                        attrs_copy[key] = str(meta_value)
            else:
                # Here we don't care about case. Within the Datasource we'll store the
                # metadata as all lowercase, within the attributes we'll keep the case.                err_warn_str = f"Metadata mismatch for '{key}', metadata: {meta_value} - attributes: {attr_value}"
                err_warn_str = (
                    f"Metadata mismatch for '{key}', metadata: {meta_value} - attributes: {attr_value}"
                )
                if str(meta_value).lower() != str(attr_value).lower():
                    if update_mismatch == "never":
                        attr_mismatches[key] = (meta_value, attr_value)
                    elif update_mismatch == "attributes":
                        logger.warning(
                            f"{err_warn_str}\nUpdating metadata to use attribute value of {key} = {attr_value}"
                        )
                        meta_copy[key] = attr_value
                    elif update_mismatch == "metadata":
                        logger.warning(
                            f"{err_warn_str}\nUpdating attributes to use metadata value: {key} = {meta_value}"
                        )
                        attrs_copy[key] = meta_value
        except KeyError:
            # Key wasn't in attributes for comparison
            pass

    if attr_mismatches:
        mismatch_details = [
            f" - '{key}', metadata: {values[0]}, attributes: {values[1]}"
            for key, values in attr_mismatches.items()
        ]
        mismatch_str = "\n".join(mismatch_details)
        raise AttrMismatchError(
            f"Metadata mismatch / value not within tolerance for the following keys:\n{mismatch_str}"
        )

    default_keys_to_add = metadata_default_keys()
    keys_as_floats = metadata_keys_as_floats()

    if keys_to_add is None:
        keys_to_add = default_keys_to_add

    # Check set of keys which should be in metadata and add if not present
    for key in keys_to_add:
        if key not in meta_copy.keys():
            try:
                meta_copy[key] = attributes[key]
            except KeyError:
                logger.warning(f"{key} key not in attributes or metadata")
            else:
                if key in keys_as_floats:
                    meta_copy[key] = float(meta_copy[key])

    return meta_copy, attrs_copy



from collections import abc
from dataclasses import dataclass
from json import dumps
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Iterator, Union, Optional
from openghg.plotting import plot_timeseries as general_plot_timeseries
import plotly.graph_objects as go

from ._basedata import _BaseData

__all__ = ["ObsData"]


@dataclass(frozen=True)
class ObsData(_BaseData, abc.Mapping):
    """This class is used to return observations data from the get_observations function

    Args:
        data: Dictionary of xarray Dataframes
        metadata: Dictionary of metadata
    """

    # Compatability layer for legacy format - mimicking the behaviour of a dictionary
    # Previous format expected a dictionary containing the site code and data
    # as key:value pairs.
    # TODO: May also want to check other expected keys within the legacy
    # dictionary format and add them below
    def __getitem__(self, key: str) -> Any:
        """
        Returns the data attribute (xarray Dataset) when the site name is
        specified.
        Included as a compatability layer for legacy format as a dictionary
        containing a Dataset for each site code.

        key (str): Site code
        """
        site = self.metadata["site"].lower()
        if key.lower() == site:
            return self.data
        else:
            raise KeyError(f"Site '{key}' does not match to expected site '{site}'")

    def __iter__(self) -> Iterator:
        """
        Returns site code as the key for the dictionary as would be expected.
        """
        site = self.metadata["site"]
        return iter([site])

    def __len__(self) -> int:
        """
        Returns number of key values (fixed at 1 at present)
        """
        # Fixed length as 1 at the moment but may need to update if other key
        # values are added.
        return 1

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ObsData):
            return NotImplemented

        return self.data.equals(other.data) and self.metadata == other.metadata

    def to_data(self) -> Dict:
        """Creates a dictionary package of this ObsData's metadata and data.

        Returns:
            dict: Dictionary of metadata and bytes
        """
        to_transfer: Dict[str, Union[str, bytes]] = {}
        to_transfer["metadata"] = dumps(self.metadata)

        # TODO - get better bytes
        with NamedTemporaryFile() as tmpfile:
            self.data.to_netcdf(tmpfile.name)
            to_transfer["data"] = Path(tmpfile.name).read_bytes()

        return to_transfer

    def plot_timeseries(
        self,
        title: Optional[str] = None,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
        units: Optional[str] = None,
        logo: Optional[bool] = True,
    ) -> go.Figure:
        """Plot a timeseries"""

        return general_plot_timeseries(
            data=self,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            units=units,
            logo=logo,
        )



from typing import Dict

import pandas as pd
from openghg.types import SurfaceTypes, optionalPathType
from openghg.util import get_datapath, get_site_info, sites_in_network

# from openghg.types import DataTypes  # Would this be more appropriate?
# This does include Footprint as well as the input obs data types?


def _extract_site_names(site_codes: list, site_filepath: optionalPathType = None) -> list:
    """
    Extracts long names for site codes.

    This uses the data stored within openghg_defs/site_info JSON file by default.

    Args:
        site_codes: List of site codes
        site_filepath: Alternative site info file.

    Returns:
        list: Long names for each site code
    """

    # Get data for site
    site_data = get_site_info(site_filepath)

    # Extracts long name from site data
    site_names = []
    for site in site_codes:
        site_details = site_data[site]
        data0 = list(site_details.values())[0]  # Uses first network entry
        try:
            site_name = data0["long_name"]
        except KeyError:
            site_name = ""
        site_names.append(site_name)

    return site_names


def summary_source_formats() -> pd.DataFrame:
    """
    Create summary DataFrame of accepted input source formats. This includes
    the site code, long name, platform and source format.

    Returns:
        pandas.DataFrame

    TODO: Add source_format details for mobile / column etc. when added
    TODO: Consider if we need to / how best to incorporate BEA2CON data
    """
    # Could include input for surface / mobile / column etc.?
    surface_source_formats = list(SurfaceTypes.__members__)

    collated_site_data = pd.DataFrame()

    # Get additional data about source formats
    # - expect details in here to be related to sites where source_format
    #   does not match to network name.
    source_format_file = get_datapath("source_format_data.csv")
    source_format_data = pd.read_csv(source_format_file)

    for source_format in surface_source_formats:
        source_format_site = source_format_data[source_format_data["source_format"] == source_format]
        site_codes = source_format_site["Site"].values

        # If no data in source format file, assume source_format may be
        # applicable to network sites.
        # TODO: May be a bad assumption, consider this and update as necessary.
        if len(site_codes) == 0:
            site_codes = sites_in_network(network=source_format)

        site_names = _extract_site_names(site_codes=site_codes)

        if len(source_format_site) > 0:
            site_data = source_format_site.copy()
        else:
            site_data = pd.DataFrame({"source_format": source_format, "Site": site_codes})
            site_data = site_data.reindex(source_format_site.columns, axis=1)

        site_data = site_data.rename(columns={"Site": "Site code", "source_format": "Source format"})
        site_data["Long name"] = site_names
        site_data["Platform"] = "surface site"

        collated_site_data = pd.concat([collated_site_data, site_data], ignore_index=True)

    # TODO: May want to sort by the site code but removed for now as nice
    # to keep GCWERKS and CRDS first.
    # collated_site_data = collated_site_data.sort_values(by="Site code")

    # Reorder columns to make "Source format" first
    columns = list(collated_site_data.columns)
    columns.remove("Source format")
    columns = ["Source format"] + columns
    collated_site_data = collated_site_data[columns]

    return collated_site_data


def summary_site_codes(site_filepath: optionalPathType = None) -> pd.DataFrame:
    """
    Create summary DataFrame of site codes. This includes details of the network,
    longitude, latitude, height above sea level and stored heights.

    Note: there may be multiple entries for the same site code if this is
    associated with multiple networks.

    Returns:
        pandas.DataFrame
    """

    # Get data for site
    site_data = get_site_info(site_filepath)

    site_dict: Dict[str, list] = {}
    site_dict["site"] = []
    site_dict["network"] = []

    expected_keys: list = ["long_name", "latitude", "longitude", "height_station_masl", ["heights", "height"]]

    name_keys = [key[0] if isinstance(key, list) else key for key in expected_keys]
    for key in name_keys:
        site_dict[key] = []

    for site, network_data in site_data.items():
        for network, data in network_data.items():
            for key in expected_keys:
                if not isinstance(key, list):
                    search_keys = [
                        key,
                    ]
                else:
                    search_keys = key

                name = search_keys[0]
                for key in search_keys:
                    if key in data:
                        site_dict[name].append(str(data[key]))
                        break
                else:
                    site_dict[name].append("")

            site_dict["network"].append(network)
            site_dict["site"].append(site)

    site_df = pd.DataFrame(site_dict)

    all_keys = name_keys.copy()
    all_keys.extend(["site", "network"])
    descriptive_names = {
        "site": "Site Code",
        "long_name": "Long name",
        "height_station_masl": "Station height (masl)",
        "heights": "Inlet heights",
    }
    column_names = {
        name: (descriptive_names[name] if name in descriptive_names else name.capitalize())
        for name in all_keys
    }

    site_df = site_df.rename(columns=column_names)
    site_df = site_df.set_index("Site Code")

    return site_df



from typing import Optional


class OpenGHGError(Exception):
    """Top level OpenGHG error"""


class InvalidSiteError(OpenGHGError):
    """Raised if an invalid site is passed"""


class UnknownDataError(OpenGHGError):
    """Raised if an unknown data type is passed"""


class FunctionError(OpenGHGError):
    """Raised if a serverless function cannot be called correctly"""


class ObjectStoreError(OpenGHGError):
    """Raised if an error accessing an object at a key in the object store occurs"""


class DatasourceLookupError(OpenGHGError):
    """Raised if Datasource lookup fails"""


class DatasourceCombineError(OpenGHGError):
    """Raised if we are unable to combine previous data within a Datasource with new data"""


class EncodingError(ObjectStoreError):
    pass


class MutexTimeoutError(OpenGHGError):
    pass


class RequestBucketError(OpenGHGError):
    pass


class SearchError(OpenGHGError):
    """Related to searching the object store"""


class AttrMismatchError(OpenGHGError):
    """Mismatch between attributes of input file and derived metadata"""


class ConfigFileError(OpenGHGError):
    """Raised for errors with configuraion file"""


class MetastoreError(OpenGHGError):
    """Raised for errors with the metadata store"""


def construct_xesmf_import_error(exception: Optional[ImportError] = None) -> str:
    xesmf_error_message = (
        "Unable to import xesmf for use with regridding algorithms."
        " To use transform modules please follow instructions"
        " for installing non-python dependencies (requires conda"
        " to be installed even if using pip to install other packages)."
    )
    # TODO: Add explicit link to instruction page once created

    if exception:
        xesmf_error_message = f"{xesmf_error_message} Full error returned: {exception}"

    return xesmf_error_message



from ._tutorial import (
    clear_tutorial_store,
    populate_flux_data,
    populate_footprint_data,
    populate_surface_data,
    populate_flux_ch4,
    populate_flux_co2,
    populate_bc_ch4,
    populate_footprint_co2,
    populate_footprint_inert,
    download_edgar_data,
    retrieve_example_data,
    retrieve_example_obspack,
    clear_example_cache,
    example_extract_path,
    use_tutorial_store,
    unpack_example_archive,
)



""" Utility functions that are used by multiple modules

"""
from collections.abc import Iterable
from typing import Any, Dict, Iterator, Optional, Tuple
import logging

logger = logging.getLogger("openghg.util")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


def running_in_cloud() -> bool:
    """Are we running in the cloud?

    Checks for the OPENGHG_CLOUD environment variable being set

    Returns:
        bool: True if running in cloud
    """
    from os import environ

    cloud_env = environ.get("OPENGHG_CLOUD", "0")

    return bool(int(cloud_env))


def running_on_hub() -> bool:
    """Are we running on the OpenGHG Hub?

    Checks for the OPENGHG_CLOUD environment variable being set

    Returns:
        bool: True if running in cloud
    """
    from os import environ

    hub_env = environ.get("OPENGHG_HUB", "0")

    return bool(int(hub_env))


def running_locally() -> bool:
    """Are we running OpenGHG locally?

    Returns:
        bool: True if running locally
    """
    return not (running_on_hub() or running_in_cloud())


def unanimous(seq: Dict) -> bool:
    """Checks that all values in an iterable object
    are the same

    Args:
        seq: Iterable object
    Returns
        bool: True if all values are the same

    """
    it = iter(seq.values())
    try:
        first = next(it)
    except StopIteration:
        return True
    else:
        return all(i == first for i in it)


def pairwise(iterable: Iterable) -> Iterator[Tuple[Any, Any]]:
    """Return a zip of an iterable where a is the iterable
    and b is the iterable advanced one step.

    Args:
        iterable: Any iterable type
    Returns:
        tuple: Tuple of iterables
    """
    from itertools import tee

    a, b = tee(iterable)
    next(b, None)

    return zip(a, b)


def site_code_finder(site_name: str) -> Optional[str]:
    """Find the site code for a given site name.

    Args:
        site_name: Site long name
    Returns:
        str or None: Three letter site code if found
    """
    from openghg.util import remove_punctuation
    from rapidfuzz import process  # type: ignore

    site_name = remove_punctuation(site_name)

    inverted = _create_site_lookup_dict()

    matches = process.extract(query=site_name, choices=inverted.keys())
    highest_score = matches[0][1]

    if highest_score < 90:
        return None

    # If there are multiple >= 90 matches we return None as this is ambiguous
    greater_than_90 = sum(match[1] >= 90 for match in matches)
    if greater_than_90 > 1:
        logger.warning("Please provide more site information, more than one site found.")
        return None

    matched_site = matches[0][0]
    site_code: str = inverted[matched_site]

    return site_code.lower()


def find_matching_site(site_name: str, possible_sites: Dict) -> str:
    """Try and find a similar name to site_name in site_list and return a suggestion or
    error string.

    Args:
        site_name: Name of site
        site_list: List of sites to check
    Returns:
        str: Suggestion / error message
    """
    from rapidfuzz import process

    site_list = possible_sites.keys()

    matches = process.extract(site_name, site_list)

    scores = [s for m, s, _ in matches]

    # This seems like a decent cutoff score for a decent find
    cutoff_score = 85

    if scores[0] < cutoff_score:
        return f"No suggestion for {site_name}."
    elif scores[0] > cutoff_score and scores[0] > scores[1]:
        best_match = matches[0][0]
        return f"Did you mean {best_match.upper()}, code: {possible_sites[best_match]} ?"
    elif scores[0] == scores[1]:
        suggestions = [f"{match.title()}, code: {possible_sites[match]}" for match, _, _ in matches]
        nl_char = "\n"
        return f"Did you mean one of : \n {nl_char.join(suggestions)}"
    else:
        return f"Unknown site: {site_name}"


def _create_site_lookup_dict() -> Dict:
    """Create a dictionary of site name: three letter site code values

    Returns:
        dict: Dictionary of site_name: site_code values
    """
    from openghg_defs import site_info_file
    from openghg.util import load_json, remove_punctuation

    site_info = load_json(path=site_info_file)

    inverted = {}
    for site, site_data in site_info.items():
        for _, network_data in site_data.items():
            try:
                long_name = network_data["long_name"]
            except KeyError:
                pass
            else:
                # Remove the country from the name
                try:
                    no_country = remove_punctuation(long_name.split(",")[0])
                except IndexError:
                    no_country = remove_punctuation(long_name)

                inverted[no_country] = site

            break

    return inverted


def verify_site(site: str) -> Optional[str]:
    """Check if the passed site is a valid one and returns the three
    letter site code if found. Otherwise we use fuzzy text matching to suggest
    sites with similar names.

    Args:
        site: Three letter site code or site name
    Returns:
        str: Verified three letter site code if valid site
    """
    from openghg.util import load_json
    from openghg_defs import site_info_file

    site_data = load_json(path=site_info_file)

    if site.upper() in site_data:
        return site.lower()
    else:
        site_code = site_code_finder(site_name=site)
        if site_code is None:
            logger.warning(f"Unable to find site code for {site}, please provide additional metadata.")
        return site_code


def multiple_inlets(site: str) -> bool:
    """Check if the passed site has more than one inlet

    Args:
        site: Three letter site code
    Returns:
        bool: True if multiple inlets
    """
    from openghg.util import get_site_info

    site_data = get_site_info()

    site = site.upper()
    network = next(iter(site_data[site]))

    try:
        heights = set(site_data[network]["height"])
    except KeyError:
        try:
            heights = set(site_data[network]["height_name"])
        except KeyError:
            return True

    return len(heights) > 1



from __future__ import annotations

from collections import defaultdict
import logging
from typing import Optional

from openghg.objectstore import get_readable_buckets
from openghg.objectstore.metastore import open_metastore
from openghg.store.spec import define_data_types
from openghg.types import ObjectStoreError


logger = logging.getLogger("openghg.objectstore")
logger.setLevel(level=logging.DEBUG)


def integrity_check(raise_error: bool = True) -> Optional[dict[str, dict[str, str]]]:
    """Check the integrity of object stores.

    Args:
        raise_error: if True, raise ObjectStoreError if integrity check fails.
        Otherwise, return a list of Datasources that failed the integrity check are returned.

    Returns:
        Nested dictionaries of failed datasource UUIDs, keyed by bucket and data type, if failures
        found, otherwise None.
    """
    from openghg.store.base import Datasource

    # For now loop over each of the object stores, can we somehow lock the object store?
    readable_buckets = get_readable_buckets()
    data_types = define_data_types()

    failed_datasources = defaultdict(dict)
    for bucket in readable_buckets.values():
        for data_type in data_types:
            # Now load the object
            with open_metastore(bucket=bucket, data_type=data_type) as metastore:
                # Get all the Datasources
                datasource_uuids = metastore.select("uuid")
                # Check they all exist
                failures = []
                for uid in datasource_uuids:
                    try:
                        Datasource.load(bucket=bucket, uuid=uid, shallow=True).integrity_check()
                    except ObjectStoreError:
                        failures.append(uid)

                if failures:
                    failed_datasources[bucket][data_type] = failures

    if failed_datasources:
        if raise_error:
            raise ObjectStoreError(
                "The following Datasources failed their integriy checks:"
                + f"\n{failed_datasources}"
                + "\nYour object store is corrupt. Please remove these Datasources."
            )
        return failed_datasources
    return None



from ._enum import (
    SurfaceTypes,
    ColumnTypes,
    ColumnSources,
    ObsTypes,
    EmissionsTypes,
    EmissionsDatabases,
    DataTypes,
)
from ._errors import (
    OpenGHGError,
    InvalidSiteError,
    UnknownDataError,
    FunctionError,
    ObjectStoreError,
    DatasourceLookupError,
    DatasourceCombineError,
    EncodingError,
    MutexTimeoutError,
    RequestBucketError,
    SearchError,
    AttrMismatchError,
    ConfigFileError,
    MetastoreError,
    construct_xesmf_import_error,
)
from ._types import (
    multiPathType,
    pathType,
    optionalPathType,
    resultsType,
    ArrayLike,
    ArrayLikeMatch,
    XrDataLike,
    XrDataLikeMatch,
)



import json
import logging
from io import BytesIO
from typing import Any, Dict, List, Optional, Union
from openghg.types import SearchError

from openghg.dataobjects import (
    BoundaryConditionsData,
    FluxData,
    FootprintData,
    ObsColumnData,
    ObsData,
)
from openghg.util import decompress, decompress_str, hash_bytes, running_on_hub
from pandas import Timestamp
from xarray import Dataset, load_dataset

logger = logging.getLogger("openghg.retrieve")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler

DataTypes = Union[BoundaryConditionsData, FluxData, FootprintData, ObsColumnData, ObsData]
multDataTypes = Union[
    List[BoundaryConditionsData], List[FluxData], List[FootprintData], List[ObsColumnData], List[ObsData]
]


def _get_generic(
    sort: bool = True,
    elevate_inlets: bool = False,
    ambig_check_params: Optional[list] = None,
    **kwargs: Any,
) -> Any:
    """Perform a search and create a dataclass object with the results if any are found.

    Args:
        data_class: Type of dataobject to create
        sort: Sort Dataset during recombination
        elevate_inlets: Elevate the inlet attribute to be a variable within the Dataset
        ambig_check_params: Parameters to check and print if result is ambiguous.
        kwargs: Additional search terms
    Returns:
        dataclass
    """
    from openghg.retrieve import search
    from openghg.store.spec import define_data_type_classes

    results = search(**kwargs)

    keyword_string = _create_keyword_string(**kwargs)
    if not results:
        err_msg = f"Unable to find results for {keyword_string}"
        logger.exception(err_msg)
        raise SearchError(err_msg)

    # TODO: UPDATE THIS - just use retrieve when retrieve_all is removed.
    retrieved_data: Any = results.retrieve_all(sort=sort, elevate_inlet=elevate_inlets)

    if retrieved_data is None:
        err_msg = f"Unable to retrieve results for {keyword_string}"
        logger.exception(err_msg)
        raise SearchError(err_msg)
    elif isinstance(retrieved_data, list) and len(retrieved_data) > 1:
        param_diff_formatted = _metadata_difference_formatted(data=retrieved_data, params=ambig_check_params)
        err_msg = f"""
        Multiple entries found for input parameters for {keyword_string}.
        Parameter differences:
        {param_diff_formatted}
        Please supply additional parameters or set ranking.
        """
        logger.exception(err_msg)
        raise SearchError(err_msg)
    elif isinstance(retrieved_data, list):
        retrieved_data = retrieved_data[0]

    # TODO: Included output of this as Any for now because we there are many
    # options for types returned but can update this

    # We can only get a single data object back here but mypy doesn't understand that
    # retrieved_data = cast(ObsData, retrieved_data)

    return retrieved_data


def get_obs_surface(
    site: str,
    species: str,
    inlet: Optional[str] = None,
    height: Optional[str] = None,
    start_date: Optional[Union[str, Timestamp]] = None,
    end_date: Optional[Union[str, Timestamp]] = None,
    average: Optional[str] = None,
    network: Optional[str] = None,
    instrument: Optional[str] = None,
    calibration_scale: Optional[str] = None,
    keep_missing: bool = False,
    skip_ranking: bool = False,
    **kwargs: Any,
) -> Optional[ObsData]:
    """This is the equivalent of the get_obs function from the ACRG repository.

    Usage and return values are the same whilst implementation may differ.

    Args:
        site: Site of interest e.g. MHD for the Mace Head site.
        species: Species identifier e.g. ch4 for methane.
        start_date: Output start date in a format that Pandas can interpret
        end_date: Output end date in a format that Pandas can interpret
        inlet: Inlet height above ground level in metres
        height: Alias for inlet
        average: Averaging period for each dataset. Each value should be a string of
        the form e.g. "2H", "30min" (should match pandas offset aliases format).
        keep_missing: Keep missing data points or drop them.
        network: Network for the site/instrument (must match number of sites).
        instrument: Specific instrument for the sipte (must match number of sites).
        calibration_scale: Convert to this calibration scale
        kwargs: Additional search terms
    Returns:
        ObsData or None: ObsData object if data found, else None
    """
    from openghg.cloud import call_function
    from openghg.util import format_inlet

    # Allow height to be an alias for inlet but we do not expect height
    # to be within the metadata (for now)
    if inlet is None and height is not None:
        inlet = height
    inlet = format_inlet(inlet)

    if running_on_hub():
        to_post: Dict[str, Union[str, Dict]] = {}

        to_post["function"] = "get_obs_surface"

        search_terms = {
            "site": site,
            "species": species,
            "keep_missing": keep_missing,
            "skip_ranking": skip_ranking,
        }

        if inlet is not None:
            search_terms["inlet"] = inlet
        if start_date is not None:
            search_terms["start_date"] = start_date
        if end_date is not None:
            search_terms["end_date"] = end_date
        if average is not None:
            search_terms["average"] = average
        if network is not None:
            search_terms["network"] = network
        if instrument is not None:
            search_terms["instrument"] = instrument
        if calibration_scale is not None:
            search_terms["calibration_scale"] = calibration_scale

        to_post["search_terms"] = search_terms

        result = call_function(data=to_post)

        content = result["content"]
        found = content["found"]

        if found:
            binary_data = decompress(data=content["data"])

            file_metadata = content["file_metadata"]
            sha1_hash_data = file_metadata["data"]["sha1_hash"]

            if sha1_hash_data != hash_bytes(data=binary_data):
                raise ValueError("Hash mismatch between local SHA1 and remote SHA1.")

            buf = BytesIO(binary_data)
            json_str = decompress_str(data=content["metadata"])
            metadata = json.loads(json_str)
            dataset = load_dataset(buf)

            return ObsData(data=dataset, metadata=metadata)
        else:
            return None
    else:
        return get_obs_surface_local(
            site=site,
            species=species,
            start_date=start_date,
            end_date=end_date,
            inlet=inlet,
            average=average,
            network=network,
            instrument=instrument,
            calibration_scale=calibration_scale,
            keep_missing=keep_missing,
            skip_ranking=skip_ranking,
            **kwargs,
        )


def get_obs_surface_local(
    site: str,
    species: str,
    inlet: Optional[str] = None,
    height: Optional[str] = None,
    start_date: Optional[Union[str, Timestamp]] = None,
    end_date: Optional[Union[str, Timestamp]] = None,
    average: Optional[str] = None,
    network: Optional[str] = None,
    instrument: Optional[str] = None,
    calibration_scale: Optional[str] = None,
    keep_missing: Optional[bool] = False,
    skip_ranking: Optional[bool] = False,
    **kwargs: Any,
) -> Optional[ObsData]:
    """This is the equivalent of the get_obs function from the ACRG repository.

    Usage and return values are the same whilst implementation may differ.

    This function should not be used on the OpenGHG Hub.

    Args:
        site: Site of interest e.g. MHD for the Mace Head site.
        species: Species identifier e.g. ch4 for methane.
        start_date: Output start date in a format that Pandas can interpret
        end_date: Output end date in a format that Pandas can interpret
        inlet: Inlet height above ground level in metres
        height: Alias for inlet
        average: Averaging period for each dataset. Each value should be a string of
        the form e.g. "2H", "30min" (should match pandas offset aliases format).
        keep_missing: Keep missing data points or drop them.
        network: Network for the site/instrument (must match number of sites).
        instrument: Specific instrument for the sipte (must match number of sites).
        calibration_scale: Convert to this calibration scale
        kwargs: Additional search terms
    Returns:
        ObsData or None: ObsData object if data found, else None
    """
    import numpy as np
    from openghg.retrieve import search_surface
    from openghg.util import clean_string, format_inlet, load_json, synonyms, timestamp_tzaware, get_site_info
    from pandas import Timedelta

    if running_on_hub():
        raise ValueError(
            "This function cannot be used on the OpenGHG Hub. Please use openghg.retrieve.get_obs_surface instead."
        )

    data_type = "surface"

    # Allow height to be an alias for inlet but we do not expect height
    # to be within the metadata (for now)
    if inlet is None and height is not None:
        inlet = height
    inlet = format_inlet(inlet)

    site_data = get_site_info()
    site = site.upper()

    # TODO: Evaluate this constraint - how do we want to handle and incorporate new sites?
    if site not in site_data:
        raise ValueError(f"No site called {site}, please enter a valid site name.")

    surface_keywords = {
        "site": site,
        "species": species,
        "inlet": inlet,
        "start_date": start_date,
        "end_date": end_date,
        "network": network,
        "instrument": instrument,
        "data_type": data_type,
    }
    surface_keywords.update(kwargs)

    # # Get the observation data
    # obs_results = search_surface(**surface_keywords)
    retrieved_data = _get_generic(
        sort=True,
        elevate_inlets=False,
        ambig_check_params=["inlet", "network", "instrument"],
        **surface_keywords,  # type: ignore
    )

    data = retrieved_data.data

    if data.attrs["inlet"] == "multiple":
        data.attrs["inlet_height_magl"] = "multiple"
        retrieved_data.metadata["inlet"] = "multiple"

    if start_date is not None and end_date is not None:
        # Check if underlying data is timezone aware.
        data_time_index = data.indexes["time"]
        tzinfo = data_time_index.tzinfo

        if tzinfo:
            start_date_filter = timestamp_tzaware(start_date)
            end_date_filter = timestamp_tzaware(end_date)
        else:
            start_date_filter = Timestamp(start_date)
            end_date_filter = Timestamp(end_date)

        end_date_filter_exclusive = end_date_filter - Timedelta(
            1, unit="nanosecond"
        )  # Deduct 1 ns to make the end day (date) exclusive.

        # Slice the data to only cover the dates we're interested in
        data = data.sel(time=slice(start_date_filter, end_date_filter_exclusive))

    try:
        start_date_data = timestamp_tzaware(data.time[0].values)
        end_date_data = timestamp_tzaware(data.time[-1].values)
    except AttributeError:
        raise AttributeError("This dataset does not have a time attribute, unable to read date range")
    except IndexError:
        return None

    if average is not None:
        # GJ - 2021-03-09
        # TODO - check by RT

        # # Average the Dataset over a given period
        # if keep_missing is True:
        #     # Create a dataset with one element and NaNs to prepend or append
        #     ds_single_element = data[{"time": 0}]

        #     for v in ds_single_element.variables:
        #         if v != "time":
        #             ds_single_element[v].values = np.nan

        #     ds_concat = []

        #     # Pad with an empty entry at the start date
        #     if timestamp_tzaware(data.time.min()) > start_date:
        #         ds_single_element_start = ds_single_element.copy()
        #         ds_single_element_start.time.values = Timestamp(start_date)
        #         ds_concat.append(ds_single_element_start)

        #     ds_concat.append(data)

        #     # Pad with an empty entry at the end date
        #     if data.time.max() < Timestamp(end_date):
        #         ds_single_element_end = ds_single_element.copy()
        #         ds_single_element_end.time.values = Timestamp(end_date) - Timedelta("1ns")
        #         ds_concat.append(ds_single_element_end)

        #     data = xr_concat(ds_concat, dim="time")

        #     # Now sort to get everything in the right order
        #     data = data.sortby("time")

        # First do a mean resample on all variables
        ds_resampled = data.resample(time=average).mean(skipna=False, keep_attrs=True)
        # keep_attrs doesn't seem to work for some reason, so manually copy
        ds_resampled.attrs = data.attrs.copy()

        average_in_seconds = Timedelta(average).total_seconds()
        ds_resampled.attrs["averaged_period"] = average_in_seconds
        ds_resampled.attrs["averaged_period_str"] = average

        # For some variables, need a different type of resampling
        data_variables: List[str] = [str(v) for v in data.variables]

        for var in data_variables:
            if "repeatability" in var:
                ds_resampled[var] = (
                    np.sqrt((data[var] ** 2).resample(time=average).sum())
                    / data[var].resample(time=average).count()
                )

            # Copy over some attributes
            if "long_name" in data[var].attrs:
                ds_resampled[var].attrs["long_name"] = data[var].attrs["long_name"]

            if "units" in data[var].attrs:
                ds_resampled[var].attrs["units"] = data[var].attrs["units"]

        # Create a new variability variable, containing the standard deviation within the resampling period
        ds_resampled[f"{species}_variability"] = (
            data[species].resample(time=average).std(skipna=False, keep_attrs=True)
        )

        # If there are any periods where only one measurement was resampled, just use the median variability
        ds_resampled[f"{species}_variability"][ds_resampled[f"{species}_variability"] == 0.0] = ds_resampled[
            f"{species}_variability"
        ].median()

        # Create attributes for variability variable
        if "long_name" in data[species].attrs:
            ds_resampled[f"{species}_variability"].attrs[
                "long_name"
            ] = f"{data[species].attrs['long_name']}_variability"

        if "units" in data[species].attrs:
            ds_resampled[f"{species}_variability"].attrs["units"] = data[species].attrs["units"]

        # Resampling may introduce NaNs, so remove, if not keep_missing
        if keep_missing is False:
            ds_resampled = ds_resampled.dropna(dim="time")

        data = ds_resampled

    # Rename variables
    rename: Dict[str, str] = {}

    data_variables = [str(v) for v in data.variables]
    for var in data_variables:
        if var.lower() == species.lower():
            rename[var] = "mf"
        if "repeatability" in var:
            rename[var] = "mf_repeatability"
        if "variability" in var:
            rename[var] = "mf_variability"
        if "number_of_observations" in var:
            rename[var] = "mf_number_of_observations"
        if "status_flag" in var:
            rename[var] = "status_flag"
        if "integration_flag" in var:
            rename[var] = "integration_flag"

    data = data.rename_vars(rename)  # type: ignore

    data.attrs["species"] = species

    if "calibration_scale" in data.attrs:
        data.attrs["scale"] = data.attrs.pop("calibration_scale")

    if calibration_scale is not None:
        data = _scale_convert(data, species, calibration_scale)

    metadata = retrieved_data.metadata
    metadata.update(data.attrs)

    obs_data = ObsData(data=data, metadata=metadata)

    # It doesn't make sense to do this now as we've only got a single Dataset
    # # Now check if the units match for each of the observation Datasets
    # units = set((f.data.mf.attrs["units"] for f in obs_files))
    # scales = set((f.data.attrs["scale"] for f in obs_files))

    # if len(units) > 1:
    #     raise ValueError(
    #         f"Units do not match for these observation Datasets {[(f.mf.attrs['station_long_name'],f.attrs['units']) for f in obs_files]}"
    #     )

    # if len(scales) > 1:
    #     print(
    #         f"Scales do not match for these observation Datasets {[(f.mf.attrs['station_long_name'],f.attrs['units']) for f in obs_files]}"
    #     )
    #     print("Suggestion: set calibration_scale to convert scales")

    return obs_data


def get_obs_column(
    species: str,
    satellite: Optional[str] = None,
    domain: Optional[str] = None,
    selection: Optional[str] = None,
    site: Optional[str] = None,
    network: Optional[str] = None,
    instrument: Optional[str] = None,
    platform: str = "satellite",
    start_date: Optional[Union[str, Timestamp]] = None,
    end_date: Optional[Union[str, Timestamp]] = None,
    **kwargs: Any,
) -> ObsColumnData:
    """
    Extract available column data from the object store using keywords.

    Args:
        species: Species name
        source: Source name
        domain: Domain e.g. EUROPE
        start_date: Start date
        end_date: End date
        time_resolution: One of ["standard", "high"]
        kwargs: Additional search terms
    Returns:
        ObsColumnData: ObsColumnData object
    """
    obs_data = _get_generic(
        sort=False,
        species=species,
        satellite=satellite,
        domain=domain,
        selection=selection,
        site=site,
        network=network,
        instrument=instrument,
        platform=platform,
        start_date=start_date,
        end_date=end_date,
        data_type="column",
        **kwargs,
    )

    return ObsColumnData(data=obs_data.data, metadata=obs_data.metadata)


def get_flux(
    species: str,
    source: str,
    domain: str,
    database: Optional[str] = None,
    database_version: Optional[str] = None,
    model: Optional[str] = None,
    start_date: Optional[Union[str, Timestamp]] = None,
    end_date: Optional[Union[str, Timestamp]] = None,
    time_resolution: Optional[str] = None,
    **kwargs: Any,
) -> FluxData:
    """
    The flux function reads in all flux files for the domain and species as an xarray Dataset.
    Note that at present ALL flux data is read in per species per domain or by emissions name.
    To be consistent with the footprints, fluxes should be in mol/m2/s.

    Args:
        species: Species name
        source: Source name
        domain: Domain e.g. EUROPE
        start_date: Start date
        end_date: End date
        time_resolution: One of ["standard", "high"]
        kwargs: Additional search terms
    Returns:
        FluxData: FluxData object
    """
    em_data = _get_generic(
        sort=False,
        species=species,
        source=source,
        domain=domain,
        database=database,
        database_version=database_version,
        model=model,
        time_resolution=time_resolution,
        start_date=start_date,
        end_date=end_date,
        data_type="emissions",
        **kwargs,
    )

    em_ds = em_data.data
    # Check for level coordinate. If one level, assume surface and drop
    if "lev" in em_ds.coords:
        if len(em_ds.lev) > 1:
            raise ValueError("Error: More than one flux level")

        em_ds = em_ds.drop_vars(names="lev")

    return FluxData(data=em_data.data, metadata=em_data.metadata)


def get_bc(
    species: str,
    domain: str,
    bc_input: Optional[str] = None,
    start_date: Optional[Union[str, Timestamp]] = None,
    end_date: Optional[Union[str, Timestamp]] = None,
    **kwargs: Any,
) -> BoundaryConditionsData:
    """
    Get boundary conditions for a given species, domain and bc_input name.

    Args:
        species: Species name
        bc_input: Input used to create boundary conditions. For example:
            - a model name such as "MOZART" or "CAMS"
            - a description such as "UniformAGAGE" (uniform values based on AGAGE average)
        domain: Region for boundary conditions e.g. EUROPE
        start_date: Start date
        end_date: End date
    Returns:
        BoundaryConditionsData: BoundaryConditionsData object
    """
    bc_data = _get_generic(
        sort=False,
        species=species,
        bc_input=bc_input,
        domain=domain,
        start_date=start_date,
        end_date=end_date,
        data_type="boundary_conditions",
        **kwargs,
    )

    return BoundaryConditionsData(data=bc_data.data, metadata=bc_data.metadata)


def get_footprint(
    site: str,
    domain: str,
    inlet: Optional[str] = None,
    height: Optional[str] = None,
    model: Optional[str] = None,
    start_date: Optional[Union[str, Timestamp]] = None,
    end_date: Optional[Union[str, Timestamp]] = None,
    species: Optional[str] = None,
    **kwargs: Any,
) -> FootprintData:
    """
    Get footprints from one site.

    Args:
        site: The name of the site given in the footprints. This often matches
              to the site name but  if the same site footprints are run with a
              different met and they are named slightly differently from the obs
              file. E.g. site="DJI", site_modifier = "DJI-SAM" -
              station called DJI, footprints site called DJI-SAM
        domain : Domain name for the footprints
        inlet: Height above ground level in metres
        height: Alias for inlet
        model: Model used to create footprint (e.g. NAME or FLEXPART)
        start_date: Output start date in a format that Pandas can interpret
        end_date: Output end date in a format that Pandas can interpret
        species: Species identifier e.g. "co2" for carbon dioxide. Only needed
                 if species needs a modified footprints from the typical 30-day
                 footprints appropriate for a long-lived species (like methane)
                 e.g. for high time resolution (co2) or is a short-lived species.
        kwargs: Additional search terms
    Returns:
        FootprintData: FootprintData dataclass
    """
    from openghg.util import clean_string, synonyms, format_inlet

    # Find the correct synonym for the passed species
    if species is not None:
        species = clean_string(synonyms(species))

    # Allow inlet or height to be specified, both or either may be included
    # within the metadata so could use either to search
    inlet = format_inlet(inlet)
    height = format_inlet(height)

    fp_data = _get_generic(
        site=site,
        domain=domain,
        inlet=inlet,
        height=height,
        model=model,
        start_date=start_date,
        end_date=end_date,
        species=species,
        data_type="footprints",
        **kwargs,
    )

    return FootprintData(data=fp_data.data, metadata=fp_data.metadata)

    # TODO: Could incorporate this somewhere? Setting species to INERT?
    # if species is None:
    #     species = metadata.get("species", "INERT")


def _scale_convert(data: Dataset, species: str, to_scale: str) -> Dataset:
    """Convert to a new calibration scale

    Args:
        data: Must contain an mf variable (mole fraction), and scale must be in global attributes
        species: species name
        to_scale: Calibration scale to convert to
    Returns:
        xarray.Dataset: Dataset with mole fraction data scaled
    """
    from numexpr import evaluate
    from openghg.util import get_datapath
    from pandas import read_csv

    # If scale is already correct, return
    ds_scale = data.attrs["scale"]
    if ds_scale == to_scale:
        return data

    scale_convert_filepath = get_datapath("acrg_obs_scale_convert.csv")

    scale_converter = read_csv(scale_convert_filepath)
    scale_converter_scales = scale_converter[scale_converter.isin([species.upper(), ds_scale, to_scale])][
        ["species", "scale1", "scale2"]
    ].dropna(axis=0, how="any")

    if len(scale_converter_scales) == 0:
        raise ValueError(
            f"Scales {ds_scale} and {to_scale} are not both in any one row in acrg_obs_scale_convert.csv for species {species}"
        )
    elif len(scale_converter_scales) > 1:
        raise ValueError("Duplicate rows in acrg_obs_scale_convert.csv?")
    else:
        row = scale_converter_scales.index[0]

    converter = scale_converter.loc[row]

    if to_scale == converter["scale1"]:
        direction = "2to1"
    else:
        direction = "1to2"

    # flake8: noqa: F841
    # scale_convert file has variable X in equations, so let's create it
    X = 1.0
    scale_factor = evaluate(converter[direction])
    data["mf"].values *= scale_factor

    data.attrs["scale"] = to_scale

    return data


def _create_keyword_string(**kwargs: Any) -> str:
    """
    Create a formatted string for supplied keyword values. This will ignore
    keywords where the value is None.
    This is used for printing details of keywords passed to the search functions.
    """
    used_keywords = {key: value for key, value in kwargs.items() if value is not None}
    keyword_string = ", ".join([f"{key}='{value}'" for key, value in used_keywords.items()])

    return keyword_string


def _metadata_difference(
    data: multDataTypes, params: Optional[list] = None, print_output: bool = True
) -> Dict[str, list]:
    """Check differences between metadata for returned data objects. Note this will
    only look at differences between values which are strings (not lists, floats etc.)

    Args:
        data: Multiple data objects e.g. multiple ObsData as a list
        params: Specific metadata parameters to check. If None all parameters will be checked
        print_output: Summarise and print output to screen.
    Returns:
        Dict[str, list]: Keys and lists of values from the metadata with differences.
    """
    # Extract metadata dictionaries from each data object in list
    metadata = [d.metadata for d in data]

    if not metadata:
        err_msg = "Unable to read metadata."
        logger.exception(err_msg)
        raise ValueError(err_msg)

    # Creating multiple metadata dictionaries to be compared
    # - Check if only selected parameters be included
    if params is not None:
        metadata = [{param: m[param] for param in params} for m in metadata]

    # - Check if some parameters should be explicitly ignored and not compared
    ignore_params = ["uuid", "data_owner", "data_owner_email"]
    if ignore_params is not None:
        metadata = [{key: value for key, value in m.items() if key not in ignore_params} for m in metadata]

    # - Extract string values  only from the underlying metadata dictionaries
    metadata = [{key: value for key, value in m.items() if isinstance(value, str)} for m in metadata]

    # Select first metadata dictionary from list and use this to compare to others
    # - Look at difference between first metadata dict and other metadata dicts
    metadata0 = metadata[0]
    difference = []
    for metadata_compare in metadata[1:]:
        try:
            metadata_diff = set(metadata0.items()) - set(metadata_compare.items())
        except TypeError:
            logger.warning("Unable to compare metadata between ambiguous results")
            return {}
        else:
            difference.extend(list(metadata_diff))
    # - Select parameter names for values which are different between metadata dictionaries
    param_difference = list(set([d[0] for d in difference]))

    # ignore_params = ["data_owner", "data_owner_email"]
    # for iparam in ignore_params:
    #     try:
    #         param_difference.remove(iparam)
    #     except ValueError:
    #         continue

    # - Collate summary of differences as a dictionary which maps as param: list of values
    summary_difference: Dict[str, list] = {}
    for param in param_difference:
        summary_difference[param] = []
        if print_output:
            logger.info(f" {param}: ")
        for m in metadata:
            if param in m:
                value = m[param]
            else:
                value = "NOT PRESENT"
            summary_difference[param].append(value)
            if print_output:
                logger.info(f" '{value}', ")
        if print_output:
            logger.info("\n")  # print new line

    # if print_output:
    #     print("Datasets contain:")
    #     for param in param_difference:
    #         print(f" {param}: ", end="")
    #         for m in metadata:
    #             print(f" '{m[param]}', ", end="")
    #         print()  # print new line

    return summary_difference


def _metadata_difference_formatted(
    data: multDataTypes, params: Optional[list] = None, print_output: bool = True
) -> str:
    """
    Create formatted string for the difference in metadata between input objects.

    Args:
        data : Multiple data objects e.g. multiple ObsData as a list
        params : Specific metadata parameters to check. If None all parameters will be checked
        print_output : Summarise and print output to screen.

    Returns:
        str : Formatted string summarising differences in keys and sets of values
              from the metadata.
    """
    param_difference = _metadata_difference(data, params, print_output)
    formatted = "\n".join([f" - {key}: {', '.join(values)}" for key, values in param_difference.items()])
    return formatted



"""
This module defines an interface for metastores.

Metastores store metadata that is used to search an
object store for data.

The documentation in this module refers to metadata entries stored
in the metastore as "records". These correspond to "documents" in
TinyDB and Mongodb.

A MetaStore managed by an ObjectStore object will store
UUIDs, which are used to locate data stored in the object store.

UUIDs are managed by ObjectStore objects, so they are not part of
the MetaStore interface.

"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union

import tinydb
from tinydb.operations import delete as tinydb_delete

from openghg.types import MetastoreError

MetaData = Dict[str, Any]
QueryResults = List[Dict[str, Any]]  # ...to avoid clashes with `SearchResults` object
Bucket = str


class MetaStore(ABC):
    """Interface for MetaStore.

    All classes implementing this interface must provide the following methods:
        search: to search for records in the metastore matching given key-value pairs
        add: to add metadata to a new record in the metastore
        delete: to delete records matching the given key-value pairs
        update: update the metadata of a record matching the `where` key-value pairs;
            keys may be added or updated (using `to_update`) or deleted (using `to_delete`)
    """

    @abstractmethod
    def search(self, search_terms: Optional[MetaData] = None) -> QueryResults:
        """Search for data using a dictionary of search terms.

        Args:
            search_terms: dictionary of key-value pairs to search by.

        Returns:
            list of records in the metastore matching the given search terms.
            If search_terms is None, return all records.
        """
        pass

    @abstractmethod
    def insert(self, metadata: MetaData) -> None:
        """Insert new metadata into the metastore."""
        pass

    @abstractmethod
    def delete(self, metadata: MetaData, delete_one: bool = True) -> None:
        """Delete metadata from the metastore.

        By default, an error will be thrown if more than one record will
        be deleted.

        If `delete_one` is False, then this will delete *all* records matching
        the given metadata. To see what will be deleted, search the metastore using
        the same metadata.

        Args:
            metadata: metadata to search for records to delete.
            delete_one: if True, throw error if more than one record will
                be deleted.

        Returns:
            None

        Raises:
            MetastoreError if delete_one=True and multiple records will
                be deleted.
        """
        if delete_one:
            if not self._uniquely_identifies(metadata):
                raise MetastoreError(
                    "Multiple records found matching metadata. "
                    "Pass `delete_one=False` to delete multiple records."
                )

    @abstractmethod
    def update(
        self,
        where: MetaData,
        to_update: Optional[MetaData] = None,
        to_delete: Optional[Union[str, list[str]]] = None,
    ) -> None:
        """Update a single record with given metadata.

        Args:
            where: metadata identifying the record to update. This must uniquely
                identify the record.
            to_update: metadata to overwrite or add to the record.
            to_delete: key or list of keys to delete from record.
        Returns:
            None

        Raises:
            MetastoreError if more than one record matches the metadata in `where`.
        """
        if not self._uniquely_identifies(where):
            raise MetastoreError(
                "Multiple records found matching metadata. `where` must identify a single record."
            )

    def select(self, key: str) -> List[Any]:
        """Select the values stored in all records for given key.

        Args:
            key: key to select.
        Returns:
            list: list of values stored at that key, over all records in the metastore.

        """
        return [result[key] for result in self.search()]

    def _uniquely_identifies(self, metadata: MetaData) -> bool:
        """Return true if the given metadata identifies a single record
        in the metastore.

        Args:
            metadata: metadata to test to see if it identifies a single record.

        Returns:
            True if a search for the given metadata returns a single result; False otherwise.
        """
        result = self.search(search_terms=metadata)
        return len(result) == 1


class TinyDBMetaStore(MetaStore):
    """MetaStore using a TinyDB database backend."""

    def __init__(self, database: tinydb.TinyDB) -> None:
        """Create TinyDBMetaStore object.

        Args:
            database: a TinyDB database.

        Returns:
            None
        """
        self._db = database

    def _format_metadata(self, metadata: MetaData) -> MetaData:
        """Convert all keys to lowercase.

        Args:
            metadata: metadata to format.
        Returns:
            formatted metadata.
        """
        return {k.lower(): v for k, v in metadata.items()}

    def _get_query(self, metadata: MetaData) -> tinydb.queries.QueryInstance:
        """Return a TinyDB query that searches for all records whose metadata
        contains the given metadata.

        Args:
            metadata: key-value pairs to search by

        Returns:
            TinyDB QueryInstance that can be used to search via the given metadata.
        """
        return tinydb.Query().fragment(self._format_metadata(metadata))

    def search(self, search_terms: Optional[MetaData] = None) -> QueryResults:
        """Search metastore using a dictionary of search terms.

        Args:
            search_terms: dictionary of key-value pairs to search by.
                For instance search_terms = {'site': 'TAC'} will find all results
                whose site is 'TAC'.

        Returns:
            list: list of records in the metastore matching the given search terms.
        """
        if not search_terms:
            search_terms = {}
        query = self._get_query(search_terms)
        return list(self._db.search(query))

    def insert(self, metadata: MetaData) -> None:
        """Add new metadata to the metastore.

        Args:
            metadata: metdata to add to the metastore.

        Returns:
            None
        """
        self._db.insert(self._format_metadata(metadata))

    def update(
        self,
        where: MetaData,
        to_update: Optional[MetaData] = None,
        to_delete: Optional[Union[str, list[str]]] = None,
    ) -> None:
        """Update a single record with given metadata.

        Args:
            where: metadata identifying the record to update. This must uniquely
                identify the record.
            to_update: metadata to overwrite or add to the record.
            to_delete: key or list of keys to delete from record.

        Returns:
            None

        Raises:
            MetastoreError if more than one record matches the metadata in `where`.
        """
        super().update(where, to_update, to_delete)  # Error handling
        query = self._get_query(where)
        if to_update:
            self._db.update(to_update, query)
        if to_delete:
            if not isinstance(to_delete, list):
                to_delete = [to_delete]
            for key in to_delete:
                self._db.update(tinydb_delete(key), query)

    def delete(self, metadata: MetaData, delete_one: bool = True) -> None:
        """Delete metadata from the metastore.

        By default, an error will be thrown if more than one record will
        be deleted.

        If `delete_one` is False, then this will delete *all* records matching
        the given metadata. To see what will be deleted, search the metastore using
        the same metadata.

        Args:
            metadata: metadata to search for records to delete.
            delete_one: if True, throw error if more than one record will
                be deleted.

        Returns:
            None

        Raises:
            MetastoreError if delete_one=True and multiple records will
                be deleted.
        """
        super().delete(metadata, delete_one)  # Error handling
        query = self._get_query(metadata)
        self._db.remove(query)



"""
This module contains the user interface for adding
data to the object store via transformations.
"""
from __future__ import annotations

from pathlib import Path
from typing import Optional, Union, Any

from openghg.objectstore import get_writable_bucket
from openghg.store.base import get_data_class


def transform_emissions_data(
    datapath: Union[str, Path],
    database: str,
    overwrite: bool = False,
    store: Optional[str] = None,
    **kwargs: Any,
) -> dict:
    """Read and transform an emissions database. This will find the appropriate
    parser function to use for the database specified. The necessary inputs
    are determined by which database is being used.

    The underlying parser functions will be of the form:
        - openghg.transform.emissions.parse_{database.lower()}
        - e.g. openghg.transform.emissions.parse_edgar()

    Args:
        datapath: Path to local copy of database archive (for now)
        database: Name of database
        overwrite: Should this data overwrite currently stored data
            which matches.
        store: name of object store to write data to.
        **kwargs: Inputs for underlying parser function for the database.
            Necessary inputs will depend on the database being parsed.

    Returns:
    """
    bucket = get_writable_bucket(name=store)
    dclass = get_data_class("emissions")

    with dclass(bucket) as dc:
        result = dc.transform_data(datapath=datapath, database=database, overwrite=overwrite, **kwargs)
    return result



from ._standardise import (
    standardise,
    standardise_bc,
    standardise_flux,
    standardise_footprint,
    standardise_column,
    standardise_surface,
    standardise_eulerian,
    standardise_from_binary_data,
)
from ._summarise import summary_source_formats, summary_site_codes



from __future__ import annotations
import logging
from pathlib import Path
from typing import DefaultDict, Dict, Optional, Sequence, Tuple, Union

import numpy as np
from pandas import Timedelta
from xarray import Dataset
import inspect
from openghg.store import DataSchema
from openghg.store.base import BaseStore
from openghg.types import multiPathType, pathType, resultsType, optionalPathType

logger = logging.getLogger("openghg.store")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


class ObsSurface(BaseStore):
    """This class is used to process surface observation data"""

    _data_type = "surface"
    _root = "ObsSurface"
    _uuid = "da0b8b44-6f85-4d3c-b6a3-3dde34f6dea1"
    _metakey = f"{_root}/uuid/{_uuid}/metastore"

    def read_data(
        self,
        binary_data: bytes,
        metadata: Dict,
        file_metadata: Dict,
        precision_data: Optional[bytes] = None,
        site_filepath: optionalPathType = None,
    ) -> Dict:
        """Reads binary data passed in by serverless function.
        The data dictionary should contain sub-dictionaries that contain
        data and metadata keys.

        This is clunky and the ObsSurface.read_file function could
        be tidied up quite a lot to be more flexible.

        Args:
            binary_data: Binary measurement data
            metadata: Metadata
            file_metadata: File metadata such as original filename
            precision_data: GCWERKS precision data
            site_filepath: Alternative site info file (see openghg/supplementary_data repository for format).
                Otherwise will use the data stored within openghg_defs/data/site_info JSON file by default.
        Returns:
            dict: Dictionary of result
        """
        from tempfile import TemporaryDirectory

        possible_kwargs = {
            "source_format",
            "network",
            "site",
            "inlet",
            "instrument",
            "sampling_period",
            "measurement_type",
            "overwrite",
            "source_format",
            "data_type",
        }

        # We've got a lot of functions that expect a file and read
        # metadata from its filename. As Acquire handled all of this behind the scenes
        # we'll create a temporary directory for now
        # TODO - add in just passing a filename to prevent all this read / write
        with TemporaryDirectory() as tmpdir:
            tmpdir_path = Path(tmpdir)

            try:
                filename = file_metadata["filename"]
            except KeyError:
                raise KeyError("We require a filename key for metadata read.")

            filepath = tmpdir_path.joinpath(filename)
            filepath.write_bytes(binary_data)

            meta_kwargs = {k: v for k, v in metadata.items() if k in possible_kwargs}

            if not meta_kwargs:
                raise ValueError("No valid metadata arguments passed, please check documentation.")

            if precision_data is None:
                result = self.read_file(filepath=filepath, **meta_kwargs)
            else:
                # We'll assume that if we have precision data it's GCWERKS
                # We don't read anything from the precision filepath so it's name doesn't matter
                precision_filepath = tmpdir_path.joinpath("precision_data.C")
                precision_filepath.write_bytes(precision_data)
                # Create the expected GCWERKS tuple
                result = self.read_file(
                    filepath=(filepath, precision_filepath),
                    site_filepath=site_filepath,
                    **meta_kwargs,
                )

        return result

    def read_file(
        self,
        filepath: multiPathType,
        source_format: str,
        network: str,
        site: str,
        inlet: Optional[str] = None,
        height: Optional[str] = None,
        instrument: Optional[str] = None,
        sampling_period: Optional[Union[Timedelta, str]] = None,
        calibration_scale: Optional[str] = None,
        measurement_type: str = "insitu",
        update_mismatch: str = "never",
        overwrite: bool = False,
        verify_site_code: bool = True,
        site_filepath: optionalPathType = None,
    ) -> Dict:
        """Process files and store in the object store. This function
            utilises the process functions of the other classes in this submodule
            to handle each data type.

        Args:
            filepath: Filepath(s)
            source_format: Data format, for example CRDS, GCWERKS
            site: Site code/name
            network: Network name
            inlet: Inlet height. Format 'NUMUNIT' e.g. "10m".
                If retrieve multiple files pass None, OpenGHG will attempt to
                extract this from the file.
            height: Alias for inlet.
            read inlets from data.
            instrument: Instrument name
            sampling_period: Sampling period in pandas style (e.g. 2H for 2 hour period, 2m for 2 minute period).
            measurement_type: Type of measurement e.g. insitu, flask
            update_mismatch: This determines how mismatches between the internal data
                  "attributes" and the supplied / derived "metadata" are handled.
                  This includes the options:
                      - "never" - don't update mismatches and raise an AttrMismatchError
                      - "from_source" / "attributes" - update mismatches based on input data (e.g. data attributes)
                      - "from_definition" / "metadata" - update mismatches based on associated data (e.g. site_info.json)
            overwrite: Overwrite previously uploaded data
                  verify_site_code: Verify the site code
                  site_filepath: Alternative site info file (see openghg/supplementary_data repository for format).
                      Otherwise will use the data stored within openghg_defs/data/site_info JSON file by default.
        Returns:
            dict: Dictionary of Datasource UUIDs

        TODO: Should "measurement_type" be changed to "platform" to align
        with ModelScenario and ObsColumn?
        """
        from collections import defaultdict
        from openghg.types import SurfaceTypes
        from openghg.util import (
            clean_string,
            format_inlet,
            hash_file,
            load_surface_parser,
            verify_site,
        )

        if not isinstance(filepath, list):
            filepath = [filepath]

        try:
            source_format = SurfaceTypes[source_format.upper()].value
        except KeyError:
            raise ValueError(f"Unknown data type {source_format} selected.")

        # Test that the passed values are valid
        # Check validity of site, instrument, inlet etc in 'site_info.json'
        # Clean the strings
        if verify_site_code:
            verified_site = verify_site(site=site)
            if verified_site is None:
                raise ValueError("Unable to validate site")
            else:
                site = verified_site
        else:
            site = clean_string(site)

        network = clean_string(network)
        instrument = clean_string(instrument)

        # Check if alias `height` is included instead of `inlet`
        if inlet is None and height is not None:
            inlet = height

        # Try to ensure inlet is 'NUM''UNIT' e.g. "10m"
        inlet = clean_string(inlet)
        inlet = format_inlet(inlet)

        sampling_period_seconds: Union[str, None] = None
        # If we have a sampling period passed we want the number of seconds
        if sampling_period is not None:
            # Check value passed is not just a number with no units
            try:
                float(sampling_period)
            except (ValueError, TypeError):
                # If this cannot be evaluated to a float assume this is correct form.
                pass
            else:
                raise ValueError(
                    f"Invalid sampling period: '{sampling_period}'. Must be specified as a string with unit (e.g. 1m for 1 minute)."
                )

            # Check string passed can be evaluated as a Timedelta object
            # and extract this in seconds.
            try:
                sampling_period_td = Timedelta(sampling_period)
            except ValueError:
                raise ValueError(
                    f"Could not evaluate sampling period: '{sampling_period}'. Must be specified as a string with valid unit (e.g. 1m for 1 minute)."
                )

            sampling_period_seconds = str(float(sampling_period_td.total_seconds()))

            # Check if sampling period has resolved to 0 seconds.
            if sampling_period_seconds == "0.0":
                raise ValueError(
                    f"Sampling period resolves to <= 0.0 seconds. Please check input: '{sampling_period}'"
                )

            # TODO: May want to add check for NaT or NaN

        # Load the data retrieve object
        parser_fn = load_surface_parser(source_format=source_format)

        results: resultsType = defaultdict(dict)

        # Create a progress bar object using the filepaths, iterate over this below
        for fp in filepath:
            if source_format == "GCWERKS":
                if not isinstance(fp, tuple):
                    raise TypeError("For GCWERKS data we expect a tuple of (data file, precision file).")

                try:
                    data_filepath = Path(fp[0])
                    precision_filepath = Path(fp[1])
                except (ValueError, TypeError):
                    raise TypeError(
                        "For GCWERKS data both data and precision filepaths must be given as a tuple."
                    )
            else:
                data_filepath = Path(fp)

            file_hash = hash_file(filepath=data_filepath)
            if file_hash in self._file_hashes and overwrite is False:
                logger.warning(
                    "This file has been uploaded previously with the filename : "
                    f"{self._file_hashes[file_hash]} - skipping."
                )
                break

            # Define required input parameters for parser function
            required_parameters = {
                "data_filepath": data_filepath,
                "site": site,
                "network": network,
                "inlet": inlet,
                "instrument": instrument,
                "sampling_period": sampling_period_seconds,
                "measurement_type": measurement_type,
                "site_filepath": site_filepath,
            }
            if source_format == "GCWERKS":
                required_parameters["precision_filepath"] = precision_filepath

            # Collect together optional parameters (not required but
            # may be accepted by underlying parser function)
            optional_parameters = {"update_mismatch": update_mismatch}
            # TODO: extend optional_parameters to include kwargs when added

            input_parameters = required_parameters.copy()

            # Find parameters that parser_fn accepts (must accept all required arguments already)
            signature = inspect.signature(parser_fn)
            fn_accepted_parameters = [param.name for param in signature.parameters.values()]

            # Check if optional parameters are present in function call and only use those which are.
            for param, param_value in optional_parameters.items():
                if param in fn_accepted_parameters:
                    input_parameters[param] = param_value
                else:
                    logger.warning(
                        f"Input: '{param}' (value: {param_value}) is not being used as part of the standardisation process."
                        f"This is not accepted by the current standardisation function: {parser_fn}"
                    )

            # Call appropriate standardisation function with input parameters
            data = parser_fn(**input_parameters)

            # Current workflow: if any species fails, whole filepath fails
            for key, value in data.items():
                species = key.split("_")[0]
                try:
                    ObsSurface.validate_data(value["data"], species=species)
                except ValueError:
                    logger.error(
                        f"Unable to validate and store data from file: {data_filepath.name}.",
                        f" Problem with species: {species}\n",
                    )
                    validated = False
                    break
            else:
                validated = True

            if not validated:
                continue

            # Alternative workflow: Would only stops certain species within a
            # file being written to the object store.
            # to_remove = []
            # for key, value in data.items():
            #     species = key.split('_')[0]
            #     try:
            #         ObsSurface.validate_data(value["data"], species=species)
            #     except ValueError:
            #         print(f"WARNING: standardised data for '{source_format}' is not in expected OpenGHG format.")
            #         print(f"Check data for {species}")
            #         print(value["data"])
            #         print("Not writing to object store.")
            #         to_remove.append(key)
            #
            # for remove in to_remove:
            #     data.pop(remove)

            required_keys = (
                "species",
                "site",
                "sampling_period",
                "station_long_name",
                "inlet",
                "instrument",
                "network",
                "source_format",
                "data_source",
                "icos_data_level",
                "data_type",
            )

            # Create Datasources, save them to the object store and get their UUIDs
            data_type = "surface"
            datasource_uuids = self.assign_data(
                data=data,
                overwrite=overwrite,
                data_type=data_type,
                required_keys=required_keys,
                min_keys=5,
            )

            results["processed"][data_filepath.name] = datasource_uuids

            # Store the hash as the key for easy searching, store the filename as well for
            # ease of checking by user
            # TODO - maybe add a timestamp to this string?
            self._file_hashes[file_hash] = data_filepath.name

            logger.info(f"Completed processing: {data_filepath.name}.")

        return dict(results)

    def read_multisite_aqmesh(
        self,
        data_filepath: pathType,
        metadata_filepath: pathType,
        network: str = "aqmesh_glasgow",
        instrument: str = "aqmesh",
        sampling_period: int = 60,
        measurement_type: str = "insitu",
        overwrite: bool = False,
    ) -> DefaultDict:
        """Read AQMesh data for the Glasgow network

        NOTE - temporary function until we know what kind of AQMesh data
        we'll be retrieve in the future.

        This data is different in that it contains multiple sites in the same file.
        """
        raise NotImplementedError(
            "This needs reworking for the new data storage method or removing as unused."
        )
        # from collections import defaultdict
        # from openghg.standardise.surface import parse_aqmesh
        # from openghg.store import assign_data
        # from openghg.util import hash_file

        # data_filepath = Path(data_filepath)
        # metadata_filepath = Path(metadata_filepath)

        # # Get a dict of data and metadata
        # processed_data = parse_aqmesh(data_filepath=data_filepath, metadata_filepath=metadata_filepath)

        # results: resultsType = defaultdict(dict)
        # for site, site_data in processed_data.items():
        #     metadata = site_data["metadata"]
        #     measurement_data = site_data["data"]

        #     file_hash = hash_file(filepath=data_filepath)

        #     if self.seen_hash(file_hash=file_hash) and overwrite is False:
        #         raise ValueError(
        #             f"This file has been uploaded previously with the filename : {self._file_hashes[file_hash]}."
        #         )
        #         break

        #     combined = {site: {"data": measurement_data, "metadata": metadata}}

        #     required_keys = (
        #         "site",
        #         "species",
        #         "inlet",
        #         "network",
        #         "instrument",
        #         "sampling_period",
        #         "measurement_type",
        #     )

        #     uuid = lookup_results[site]

        #     # Jump through these hoops until we can rework the data assignment functionality to split it out
        #     # into more sensible functions
        #     # TODO - fix the assign data function to avoid this kind of hoop jumping
        #     lookup_result = {site: uuid}

        #     # Create Datasources, save them to the object store and get their UUIDs
        #     data_type = "surface"
        #     datasource_uuids = assign_data(
        #         data_dict=combined,
        #         lookup_results=lookup_result,
        #         overwrite=overwrite,
        #         data_type=data_type,
        #     )

        #     results[site] = datasource_uuids

        #     # Record the Datasources we've created / appended to
        #     self.add_datasources(uuids=datasource_uuids, data=combined, metastore=self._metastore)

        #     # Store the hash as the key for easy searching, store the filename as well for
        #     # ease of checking by user
        #     self.set_hash(file_hash=file_hash, filename=data_filepath.name)

        # return results

    @staticmethod
    def schema(species: str) -> DataSchema:
        """
        Define schema for surface observations Dataset.

        Only includes mandatory variables
            - standardised species name (e.g. "ch4")
            - expected dimensions: ("time")

        Expected data types for variables and coordinates also included.

        Returns:
            DataSchema : Contains basic schema for ObsSurface.

        # TODO: Decide how to best incorporate optional variables
        # e.g. "ch4_variability", "ch4_number_of_observations"
        """
        from openghg.standardise.meta import define_species_label

        name = define_species_label(species)[0]

        data_vars: Dict[str, Tuple[str, ...]] = {name: ("time",)}
        dtypes = {name: np.floating, "time": np.datetime64}

        source_format = DataSchema(data_vars=data_vars, dtypes=dtypes)

        return source_format

    @staticmethod
    def validate_data(data: Dataset, species: str) -> None:
        """
        Validate input data against ObsSurface schema - definition from
        ObsSurface.schema() method.

        Args:
            data : xarray Dataset in expected format
            species: Species name

        Returns:
            None

            Raises a ValueError with details if the input data does not adhere
            to the ObsSurface schema.
        """
        data_schema = ObsSurface.schema(species)
        data_schema.validate_data(data)

    def store_data(
        self,
        data: Dict,
        overwrite: bool = False,
        force: bool = False,
        required_metakeys: Optional[Sequence] = None,
    ) -> Optional[Dict]:
        """This expects already standardised data such as ICOS / CEDA

        Args:
            data: Dictionary of data in standard format, see the data spec under
                Development -> Data specifications in the documentation
            overwrite: If True overwrite currently stored data
            force: Force adding of data even if this is identical to data stored (checked based on previously retrieved file hashes).
            required_metakeys: Keys in the metadata we should use to store this metadata in the object store
                if None it defaults to:
                {"species", "site", "station_long_name", "inlet", "instrument",
                "network", "source_format", "data_source", "icos_data_level"}
        Returns:
            Dict or None:
        """
        from openghg.util import hash_retrieved_data

        # Very rudimentary hash of the data and associated metadata
        hashes = hash_retrieved_data(to_hash=data)
        # Find the keys in data we've seen before
        if force:
            file_hashes_to_compare = set()
        else:
            file_hashes_to_compare = {next(iter(v)) for k, v in hashes.items() if k in self._retrieved_hashes}

        if len(file_hashes_to_compare) == len(data):
            logger.warning("Note: There is no new data to process.")
            return None

        keys_to_process = set(data.keys())
        if file_hashes_to_compare:
            # TODO - add this to log
            logger.warning(f"Note: We've seen {file_hashes_to_compare} before. Processing new data only.")
            keys_to_process -= file_hashes_to_compare

        to_process = {k: v for k, v in data.items() if k in keys_to_process}

        if required_metakeys is None:
            required_metakeys = (
                "species",
                "site",
                "station_long_name",
                "inlet",
                "instrument",
                "network",
                "source_format",
                "data_source",
                "icos_data_level",
            )

        # Create Datasources, save them to the object store and get their UUIDs
        data_type = "surface"
        # This adds the parsed data to new or existing Datasources by performing a lookup
        # in the metastore
        datasource_uuids = self.assign_data(
            data=to_process,
            overwrite=overwrite,
            data_type=data_type,
            required_keys=required_metakeys,
            min_keys=5,
        )

        self.store_hashes(hashes=hashes)

        return datasource_uuids

    def store_hashes(self, hashes: Dict) -> None:
        """Store hashes of data retrieved from a remote data source such as
        ICOS or CEDA. This takes the full dictionary of hashes, removes the ones we've
        seen before and adds the new.

        Args:
            hashes: Dictionary of hashes provided by the hash_retrieved_data function
        Returns:
            None
        """
        new = {k: v for k, v in hashes.items() if k not in self._retrieved_hashes}
        self._retrieved_hashes.update(new)

    def delete(self, uuid: str) -> None:
        """Delete a Datasource with the given UUID

        This function deletes both the record of the object store in he

        Args:
            uuid (str): UUID of Datasource
        Returns:
            None
        """
        from openghg.objectstore import delete_object
        from openghg.store.base import Datasource

        # Load the Datasource and get all its keys
        # iterate over these keys and delete them
        datasource = Datasource.load(bucket=self._bucket, uuid=uuid)

        data_keys = datasource.raw_keys()

        for version in data_keys:
            key_data = data_keys[version]["keys"]

            for daterange in key_data:
                key = key_data[daterange]
                delete_object(bucket=self._bucket, key=key)

        # Then delete the Datasource itself
        key = f"{Datasource._datasource_root}/uuid/{uuid}"
        delete_object(bucket=self._bucket, key=key)

        # Delete the UUID from the metastore
        self._metastore.delete({"uuid": uuid})

    def seen_hash(self, file_hash: str) -> bool:
        return file_hash in self._file_hashes

    def set_hash(self, file_hash: str, filename: str) -> None:
        self._file_hashes[file_hash] = filename



import numpy as np
import xarray as xr
from pathlib import Path
from typing import DefaultDict, Dict, List, Tuple, Union, Optional, TypeVar

pathType = Union[str, Path]
optionalPathType = Optional[pathType]
multiPathType = Union[str, Path, Tuple, List]
resultsType = DefaultDict[str, Dict]

# Create types for ndarray or xr.DataArray inputs
# Using TypeVar means - whichever type is passed in will be the one which is returned.
ArrayLike = Union[np.ndarray, xr.DataArray]
ArrayLikeMatch = TypeVar("ArrayLikeMatch", np.ndarray, xr.DataArray)
XrDataLike = Union[xr.DataArray, xr.Dataset]
XrDataLikeMatch = TypeVar("XrDataLikeMatch", xr.DataArray, xr.Dataset)



from typing import Optional, Tuple, Union, cast

import numpy as np
import xarray as xr
from numpy import ndarray

from openghg.types import construct_xesmf_import_error, ArrayLike, ArrayLikeMatch


def _getGridCC(lat: ndarray, lon: ndarray) -> Tuple[ndarray, ndarray]:
    """
    Create a cell centered meshgrid for from 1D arrays of lon, lat
    This meshgrid defines the bounds of each cell.
    """
    # dx = lon[2]-lon[1]
    # dy = lat[2]-lat[1]
    dx = np.mean(lon[1:] - lon[:-1])
    dy = np.mean(lat[1:] - lat[:-1])
    lon = np.append(lon, lon[-1] + dx)
    lat = np.append(lat, lat[-1] + dy)
    lon -= dx / 2.0
    lat -= dy / 2.0
    LON, LAT = np.meshgrid(lon, lat)
    return LON, LAT


# def _create_xesmf_grid_uniform_cc(lat: ndarray, lon: ndarray) -> xr.Dataset:
#     """
#     Creates a Dataset ready to be used by the xesmf regridder from 1D arrays
#     of latitude and longitude values.

#     This includes both the centre points and bounds of the latitude and longitude
#     grids (needed for conservative regridding methods but not for bilinear etc.)
#     """
#     LON, LAT = np.meshgrid(lon, lat)
#     LON_b, LAT_b = _getGridCC(lat, lon)

#     grid = xr.Dataset(
#         {
#             "lon": (["x", "y"], LON),
#             "lat": (["x", "y"], LAT),
#             "lon_b": (["x_b", "y_b"], LON_b),
#             "lat_b": (["x_b", "y_b"], LAT_b),
#         }
#     )

#     return grid


def _create_uniform_coords(lat: ndarray, lon: ndarray) -> xr.Dataset:
    """
    Creates a Dataset ready to be used by the xesmf regridder from 1D arrays
    of latitude and longitude values.
    """

    ds_out = xr.Dataset(
        {
            "lat": (["lat"], lat, {"units": "degrees_north"}),
            "lon": (["lon"], lon, {"units": "degrees_east"}),
        }
    )

    return ds_out


def convert_to_ndarray(array: Union[ndarray, xr.DataArray]) -> ndarray:
    """Check and extract underlying numpy array from DataArray as necessary"""

    if isinstance(array, xr.DataArray):
        values = array.values
    else:
        values = array

    return values


def regrid_uniform_cc(
    data: ArrayLikeMatch,
    lat_out: ArrayLike,
    lon_out: ArrayLike,
    lat_in: Optional[ArrayLike] = None,
    lon_in: Optional[ArrayLike] = None,
    latlon: Optional[list] = None,
    method: str = "conservative",
) -> ArrayLikeMatch:
    """
    Regrid data between two uniform, cell centered grids.
    All coordinates (lat_out, lon_out, lat_in, lon_in) should be for the centre
    of the representative cell and in degrees.

    Adapted from code written by @DTHoare

    Args:
        data: Data to be regridded.
            Data must have dimensions (lat, lon) if 2D or (time, lat, lon) if 3D.
        lat_out: 1D array for output latitude grid
        lon_out: 1D array for output longituide grid
        lat_in: 1D array for input latitude grid.
            Only used if data is a numpy array and not a DataArray
        lon_in: 1D array for input longitude grid.
            Only used if data is a numpy array and not a DataArray
        latlon: Names for latitude and longitude coordinates within data.
            Default = ["lat", "lon"]
        method: Method to use for regridding. Mainly use:
            - "conservative"
            - "conservative_normed" (ignores NaN values)
            See xesmf documentation for full list of options.

    Returns:
        ndarray / DataArray : Regridded data using specified method
    """
    try:
        import xesmf  # type:ignore
    except ImportError as e:
        raise ImportError(construct_xesmf_import_error(e))

    if latlon is None:
        latlon = ["lat", "lon"]

    if isinstance(data, xr.DataArray):
        lat_in_extracted = data[latlon[0]].values
        lon_in_extracted = data[latlon[1]].values

        if lat_in is not None:
            if not np.isclose(lat_in, lat_in_extracted):
                raise ValueError(
                    "Input for 'lat_in' should not be supplied if 'data' is a DataArray object.\n"
                    "Please check 'lat_out' have been supplied correctly as well."
                )

        if lon_in is not None:
            if not np.isclose(lon_in, lon_in_extracted):
                raise ValueError(
                    "Input for 'lon_in' should not be supplied if 'data' is a DataArray object.\n"
                    "Please check 'lon_out' has been supplied correctly as well."
                )

        lat_in = lat_in_extracted
        lon_in = lon_in_extracted

    lat_in = cast(ArrayLike, lat_in)
    lon_in = cast(ArrayLike, lon_in)

    if data.shape != (lat_in.size, lon_in.size):
        raise ValueError(
            f"Shape of input 'data' {data.shape}"
            f"does not match 'lat_in' and 'lon_in' lengths: {len(lat_in)}, {len(lon_in)}"
        )

    lat_in = convert_to_ndarray(lat_in)
    lon_in = convert_to_ndarray(lon_in)
    lat_out = convert_to_ndarray(lat_out)
    lon_out = convert_to_ndarray(lon_out)

    # 09/03/2023: Don't seem to need inputs with centre and bounding boxes for
    # uniform grid and xesmf "conservative" method anymore.
    input_grid = _create_uniform_coords(lat_in, lon_in)
    output_grid = _create_uniform_coords(lat_out, lon_out)

    regridder = xesmf.Regridder(input_grid, output_grid, method)
    regridded: ArrayLikeMatch = regridder(data)

    # # 09/03/2023: Don't need this for uniform grid anymore due to changes above
    # # but a variant may be needed for non-uniform grids e.g. tropomi data.
    # if isinstance(regridded, xr.DataArray):
    #     from scipy.sparse import coo_matrix  # type:ignore

    #     # Checking dimensions and mapping back lat_out and lon_out
    #     # May be overkill but attempting to make sure this is done correctly.
    #     # The regridded DataArray will contain dimensions of ('x', 'y') for
    #     # data, 'lat' and 'lon' coordinates e.g. lon = [[-10,-10],[0,0]]
    #     # This is to allow for the generic case where ('lat', 'lon') is not
    #     # a rectilinear (uniform) grid.
    #     # Since we are creating a uniform grid we want to flatten this and
    #     # put data back on ('lat', 'lon') dimension.
    #     regridded_stack = regridded.stack(z=("x", "y"))
    #     lat_coord = regridded_stack["lat"]
    #     lon_coord = regridded_stack["lon"]

    #     # Find coords within our lat_out and lon_out grid for our regridded output
    #     # Digitize is technically a binning operation outputting the number of the
    #     # bin the value is within so to get indicies we can subtract 1.
    #     lat_index = np.digitize(lat_coord, lat_out) - 1
    #     lon_index = np.digitize(lon_coord, lon_out) - 1

    #     # Create a sparse matrix from the flattened data and the lat, lon index values.
    #     # Reshape to our required lat_out, lon_out and output as a numpy array
    #     shape = (len(lat_out), len(lon_out))
    #     regridded_grid = coo_matrix((regridded_stack.data, (lat_index, lon_index)), shape=shape).toarray()

    #     regridded = xr.DataArray(
    #         regridded_grid,
    #         dims=("lat", "lon"),
    #         coords={"lat": lat_out, "lon": lon_out},
    #         attrs=regridded.attrs,
    #     )

    #     # # Alternative, more simplistic approach. May not be generic
    #     # regridded = regridded.drop(labels=("lat","lon"))
    #     # regridded = regridded.assign_coords(**{"x":output_lat,"y":output_lon})
    #     # regridded = regridded.rename({"x":"lat","y":"lon"})

    # # In later versions of xesmf (0.7?) this no longer seems to be needed
    # # as weight files are not stored on disk.
    # regridder.clean_weight_file()

    return regridded


# def regrid_betweenGrids(data, input_grid, output_grid, method="conservative"):
#     """
#     Regrid data from predefined input_grid and output_grid

#     Inputs
#         data - numpy array

#         input_grid, output_grid -  Dataset of the form:

#             xr.Dataset({'lat': (['x', 'y'], LAT),
#                          'lon': (['x', 'y'], LON),
#                          'lat_b': (['x_b', 'y_b'], LAT_b),
#                          'lon_b': (['x_b', 'y_b'], LON_b)})

#             where lat and lon give cell centre locations, and lat_b and lon_b give the cell bounds (corners)

#         method - string describing method to use.
#             Should be one of "conservative" or "conservative_normed".
#             Note that to use the conservative_normed method you need to have installed the "masking" development
#             branch of the xesmf package (contact Daniel for more details).

#             With the 'masking' branch of xESMF you can include a mask in the input_grid to ignore nan values
#     returns
#         regridded numpy array
#     """
#     try:
#         import xesmf
#     except ImportError:
#         raise ImportError(xesmf_error_message)

#     #if 'mask' in input_grid:
#     #    # !!! Requires the 'masking' branch of xESMF to be manually installed !!!
#     #    method = 'conservative_normed'
#     #else:
#     #method = 'conservative'
#     regridder = xesmf.Regridder(input_grid, output_grid, method)
#     regridded = regridder( data )
#     regridder.clean_weight_file()

#     return regridded



from ._access import (
    get_bc,
    get_flux,
    get_footprint,
    get_obs_column,
    get_obs_surface,
    get_obs_surface_local,
)
from ._export import get_ceda_file
from ._search import (
    search,
    search_bc,
    search_column,
    search_flux,
    search_eulerian,
    search_footprints,
    search_surface,
)



""" This file contains the BaseStore class from which other storage
    modules inherit.
"""
from __future__ import annotations

import logging
from types import TracebackType
from typing import Any, Dict, List, Optional, Sequence, TypeVar, Union

from pandas import Timestamp

from openghg.objectstore import get_object_from_json, exists, set_object_from_json
from openghg.objectstore.metastore import DataClassMetaStore
from openghg.types import DatasourceLookupError
from openghg.util import timestamp_now, to_lowercase


T = TypeVar("T", bound="BaseStore")


logger = logging.getLogger("openghg.store")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


class BaseStore:
    _registry: dict[str, type[BaseStore]] = {}
    _data_type = ""
    _root = "root"
    _uuid = "root_uuid"

    def __init__(self, bucket: str) -> None:
        self._creation_datetime = str(timestamp_now())
        self._stored = False
        # Hashes of previously uploaded files
        self._file_hashes: Dict[str, str] = {}
        # Hashes of previously stored data from other data platforms
        self._retrieved_hashes: Dict[str, Dict] = {}
        # Where we'll store this object's metastore
        self._metakey = ""

        if exists(bucket=bucket, key=self.key()):
            data = get_object_from_json(bucket=bucket, key=self.key())
            # Update myself
            self.__dict__.update(data)

        self._metastore = DataClassMetaStore(bucket=bucket, data_type=self._data_type)
        self._bucket = bucket
        self._datasource_uuids = self._metastore.select("uuid")

    def __init_subclass__(cls) -> None:
        BaseStore._registry[cls._data_type] = cls

    def __enter__(self) -> BaseStore:
        return self

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        if exc_type is not None:
            logger.error(msg=f"{exc_type}, {exc_tb}")
        else:
            self.save()

    @classmethod
    def metakey(cls) -> str:
        return str(cls._metakey)

    @classmethod
    def key(cls) -> str:
        return f"{cls._root}/uuid/{cls._uuid}"

    def save(self) -> None:
        self._metastore.close()
        set_object_from_json(bucket=self._bucket, key=self.key(), data=self.to_data())

    def to_data(self) -> Dict:
        # We don't need to store the metadata store, it has its own location
        # QUESTION - Is this cleaner than the previous specifying
        DO_NOT_STORE = ["_metastore", "_bucket", "_datasource_uuids"]
        return {k: v for k, v in self.__dict__.items() if k not in DO_NOT_STORE}

    def read_data(self, *args: Any, **kwargs: Any) -> Optional[dict]:
        raise NotImplementedError

    def read_file(self, *args: Any, **kwargs: Any) -> dict:
        raise NotImplementedError

    def store_data(self, *args: Any, **kwargs: Any) -> Optional[dict]:
        raise NotImplementedError

    def transform_data(self, *args: Any, **kwargs: Any) -> dict:
        raise NotImplementedError

    def assign_data(
        self,
        data: Dict,
        overwrite: bool,
        data_type: str,
        required_keys: Sequence[str],
        min_keys: Optional[int] = None,
        update_keys: Optional[List] = None,
    ) -> Dict[str, Dict]:
        """Assign data to a Datasource. This will either create a new Datasource
        Create or get an existing Datasource for each gas in the file

            Args:
                data: Dictionary containing data and metadata for species
                overwrite: If True overwrite current data stored
                data_type: Type of data, timeseries etc
                required_keys: Required minimum keys to lookup unique Datasource
                min_keys: Minimum number of metadata keys needed to uniquely match a Datasource
            Returns:
                dict: Dictionary of UUIDs of Datasources data has been assigned to keyed by species name
        """
        from openghg.store.base import Datasource

        uuids = {}

        self._metastore.acquire_lock()
        try:
            lookup_results = self.datasource_lookup(data=data, required_keys=required_keys, min_keys=min_keys)
            # TODO - remove this when the lowercasing of metadata gets removed
            # We currently lowercase all the metadata and some keys we don't want to change, such as paths to the object store
            skip_keys = ["object_store"]

            for key, parsed_data in data.items():
                metadata = parsed_data["metadata"]
                _data = parsed_data["data"]

                # Our lookup results and gas data have the same keys
                uuid = lookup_results[key]

                # Add the read metadata to the Dataset attributes being careful
                # not to overwrite any attributes that are already there
                def convert_to_netcdf4_types(value: Any) -> Union[int, float, str, list]:
                    """Attributes in a netCDF file can be strings, numbers, or sequences:
                    http://unidata.github.io/netcdf4-python/#attributes-in-a-netcdf-file

                    This function converts any data whose type is not int, float, str, or list
                    to strings.
                    Booleans are converted to strings, even though they are a subtype of int.
                    """
                    if isinstance(value, (int, float, str, list)) and not isinstance(value, bool):
                        return value
                    else:
                        return str(value)

                to_add = {k: convert_to_netcdf4_types(v) for k, v in metadata.items() if k not in _data.attrs}
                _data.attrs.update(to_add)

                # If we have a UUID for this Datasource load the existing object
                # from the object store
                # If we haven't stored data with this metadata before we create a new Datasource
                # and add the metadata to our metastore

                # Take a copy of the metadata so we can update it
                meta_copy = metadata.copy()
                new_ds = uuid is False

                if new_ds:
                    datasource = Datasource()
                    uid = datasource.uuid()
                    meta_copy["uuid"] = uid
                    # Make sure all the metadata is lowercase for easier searching later
                    # TODO - do we want to do this or should be just perform lowercase comparisons?
                    meta_copy = to_lowercase(d=meta_copy, skip_keys=skip_keys)
                else:
                    datasource = Datasource.load(bucket=self._bucket, uuid=uuid)

                # Add the dataframe to the datasource
                datasource.add_data(
                    metadata=meta_copy,
                    data=_data,
                    overwrite=overwrite,
                    data_type=data_type,
                    skip_keys=skip_keys,
                )
                # Save Datasource to object store
                datasource.save(bucket=self._bucket)

                # Add the metadata to the metastore and make sure it's up to date with the metadata stored
                # in the Datasource
                datasource_metadata = datasource.metadata()
                if new_ds:
                    self._metastore.insert(datasource_metadata)
                else:
                    self._metastore.update(where={"uuid": datasource.uuid()}, to_update=datasource_metadata)

                uuids[key] = {"uuid": datasource.uuid(), "new": new_ds}
        finally:
            self._metastore.release_lock()

        return uuids

    def datasource_lookup(
        self, data: Dict, required_keys: Sequence[str], min_keys: Optional[int] = None
    ) -> Dict:
        """Search the metadata store for a Datasource UUID using the metadata in data. We expect the required_keys
        to be present and will require at least min_keys of these to be present when searching.

        As some metadata value might change (such as data owners etc) we don't want to do an exact
        search on *all* the metadata so we extract a subset (the required keys) and search for these.

        Args:
            metastore: Metadata database
            data: Combined data dictionary of form {key: {data: Dataset, metadata: Dict}}
            required_keys: Iterable of keys to extract from metadata
            min_keys: The minimum number of required keys, if not given it will be set
            to the length of required_keys
        Return:
            dict: Dictionary of datasource information
        """
        from openghg.util import to_lowercase

        if min_keys is None:
            min_keys = len(required_keys)

        results = {}
        for key, _data in data.items():
            metadata = _data["metadata"]

            required_metadata = {
                k.lower(): to_lowercase(v) for k, v in metadata.items() if k in required_keys
            }

            if len(required_metadata) < min_keys:
                raise ValueError(
                    f"The given metadata doesn't contain enough information, we need: {required_keys}"
                )

            required_result = self._metastore.search(required_metadata)

            if not required_result:
                results[key] = False
            elif len(required_result) > 1:
                raise DatasourceLookupError("More than once Datasource found for metadata, refine lookup.")
            else:
                results[key] = required_result[0]["uuid"]

        return results

    def uuid(self) -> str:
        """Return the UUID of this object

        Returns:
            str: UUID of object
        """
        return self._uuid

    def datasources(self) -> List[str]:
        """Return the list of Datasources UUIDs associated with this object

        Returns:
            list: List of Datasource UUIDs
        """
        return self._datasource_uuids

    def get_rank(self, uuid: str, start_date: Timestamp, end_date: Timestamp) -> Dict:
        """Get the rank for the given Datasource for a given date range

        Args:
            uuid: UUID of Datasource
            start_date: Start date
            end_date: End date
        Returns:
            dict: Dictionary of rank and daterange covered by that rank
        """
        raise NotImplementedError("Ranking is being reworked and will be reactivated in a future release.")
        from collections import defaultdict
        from openghg.util import create_daterange_str, daterange_overlap

        if uuid not in self._rank_data:
            return {}

        search_daterange = create_daterange_str(start=start_date, end=end_date)

        rank_data = self._rank_data[uuid]

        ranked = defaultdict(list)
        # Check if this Datasource is ranked for the dates passed
        for daterange, rank in rank_data.items():
            if daterange_overlap(daterange_a=search_daterange, daterange_b=daterange):
                ranked[rank].append(daterange)

        return ranked

    def clear_rank(self, uuid: str) -> None:
        """Clear the ranking data for a Datasource

        Args:
            uuid: UUID of Datasource
        Returns:
            None
        """
        raise NotImplementedError("Ranking is being reworked and will be reactivated in a future release.")
        if uuid in self._rank_data:
            del self._rank_data[uuid]
            self.save()
        else:
            raise ValueError("No ranking data set for that UUID.")

    def set_rank(
        self,
        uuid: str,
        rank: Union[int, str],
        date_range: Union[str, List[str]],
        overwrite: Optional[bool] = False,
    ) -> None:
        """Set the rank of a Datasource associated with this object.

        This function performs checks to ensure multiple ranks aren't set for
        overlapping dateranges.

        Passing a daterange and rank to this function will overwrite any current
        daterange stored for that rank.

        Args:
            uuid: UUID of Datasource
            rank: Rank of data
            date_range: Daterange(s)
            overwrite: Overwrite current ranking data
        Returns:
            None
        """
        raise NotImplementedError("Ranking is being reworked and will be reactivated in a future release.")
        from copy import deepcopy

        from openghg.util import (
            combine_dateranges,
            daterange_contains,
            daterange_overlap,
            sanitise_daterange,
            split_encompassed_daterange,
            trim_daterange,
        )

        rank = int(rank)

        if not 1 <= rank <= 10:
            raise TypeError("Rank can only take values 1 to 10 (for unranked). Where 1 is the highest rank.")

        if not isinstance(date_range, list):
            date_range = [date_range]

        # Make sure the dateranges passed are correct and are tz-aware
        date_range = [sanitise_daterange(d) for d in date_range]
        # Combine in case we have overlappng dateranges
        date_range = combine_dateranges(date_range)

        # Used to store dateranges that need to be trimmed to ensure no daterange overlap
        to_update = []
        # Non-overlapping dateranges that can be stored directly
        to_add = []

        if uuid in self._rank_data:
            rank_data = self._rank_data[uuid]
            # Check this source isn't ranked differently for the same dates
            for new_daterange in date_range:
                overlap = False
                # Check for overlapping dateranges and add
                for existing_daterange, existing_rank in rank_data.items():
                    if daterange_overlap(daterange_a=new_daterange, daterange_b=existing_daterange):
                        overlap = True

                        if rank != existing_rank and overwrite:
                            # Save the daterange we need to update
                            to_update.append((existing_daterange, new_daterange))
                            continue
                        # If the ranks are the same we just want to combine the dateranges
                        elif rank == existing_rank:
                            to_combine = [new_daterange, existing_daterange]
                            combined = combine_dateranges(dateranges=to_combine)[0]
                            to_update.append((existing_daterange, combined))
                        else:
                            raise ValueError(
                                f"This datasource has rank {existing_rank} for dates that overlap the ones given. \
                                                Overlapping dateranges are {new_daterange} and {existing_daterange}"
                            )
                # Otherwise we just want to add the new daterange to the dict
                if not overlap:
                    to_add.append(new_daterange)

            # If we've got dateranges to update and ranks to overwrite we need to trim the
            # previous ranking daterange down so we don't have overlapping dateranges
            if to_update:
                # Here we first take a backup of the old ranking data, update
                # it and then write it back
                ranking_backup = deepcopy(rank_data)

                for existing, new in to_update:
                    # Remove the existing daterange key
                    # Here we pass if it doesn't exist as if we have multiple new dateranges
                    # that overlap the existing daterange it might have been deleted during
                    # a previous iteration
                    try:
                        del ranking_backup[existing]
                    except KeyError:
                        pass

                    # If we want to overwrite an existing rank we need to trim that daterange and
                    # rewrite it back to the dictionary
                    rank_copy = rank_data[existing]

                    if overwrite:
                        if existing == new:
                            ranking_backup[new] = rank_copy
                        # If the existing daterange contains the new daterange
                        # we need to split it into parts and save those
                        elif daterange_contains(container=existing, contained=new):
                            result = split_encompassed_daterange(container=existing, contained=new)

                            existing_start = result["container_start"]
                            ranking_backup[existing_start] = rank_copy

                            updated_new = result["contained"]
                            ranking_backup[updated_new] = rank

                            # We might only end up with two dateranges
                            try:
                                existing_end = result["container_end"]
                                ranking_backup[existing_end] = rank_copy
                            except KeyError:
                                pass
                        # If the new daterange contains the existing we can just overwrite it
                        elif daterange_contains(container=new, contained=existing):
                            ranking_backup[new] = rank
                        else:
                            trimmed = trim_daterange(to_trim=existing, overlapping=new)
                            ranking_backup[trimmed] = rank_copy
                            ranking_backup[new] = rank
                    elif rank_copy == rank:
                        # If we're not overwriting we just need to update to use the new combined
                        ranking_backup[new] = rank_copy

                self._rank_data[uuid] = ranking_backup

            # Finally, store the dateranges that didn't overlap
            for d in to_add:
                self._rank_data[uuid][d] = rank
        else:
            for d in date_range:
                self._rank_data[uuid][d] = rank

        self.save()

    def rank_data(self) -> Dict:
        """Return a dictionary of rank data keyed
        by UUID

            Returns:
                dict: Dictionary of rank data
        """
        raise NotImplementedError("Ranking is being reworked and will be reactivated in a future release.")
        rank_dict: Dict = self._rank_data.to_dict()
        return rank_dict

    def clear_datasources(self) -> None:
        """Remove all Datasources from the object

        Returns:
            None
        """
        self._datasource_uuids.clear()
        self._file_hashes.clear()


def get_data_class(data_type: str) -> type[BaseStore]:
    """Return data class corresponding to given data type.

    Args:
        data_type: one of "surface", "column", "emissions", "footprints",
    "boundary_conditions", or "eulerian_model"

    Returns:
        Data class, one of `ObsSurface`, `ObsColumn`, `Emissions`, `EulerianModel`,
    `Footprints`, `BoundaryConditions`.
    """
    try:
        data_class = BaseStore._registry[data_type]
    except KeyError:
        raise ValueError(f"No data class for data type {data_type}.")
    else:
        return data_class



from typing import Dict, Any
from openghg.util import load_json
from openghg.types import optionalPathType

__all__ = ["get_site_info", "sites_in_network"]


def get_site_info(site_filepath: optionalPathType = None) -> Dict[str, Any]:
    """Extract data from site info JSON file as a dictionary.

    This uses the data stored within openghg_defs/data/site_info JSON file by default.

    Args:
        site_filepath: Alternative site info file.
    Returns:
        dict: Data from site JSON file
    """
    from openghg_defs import site_info_file

    if site_filepath is None:
        site_info_json = load_json(path=site_info_file)
    else:
        site_info_json = load_json(path=site_filepath)

    return site_info_json


def sites_in_network(network: str, site_filepath: optionalPathType = None) -> list:
    """Extract details of all the sites within a network.
    Note: this will assume the network is stored in upper case.

    Args:
        network: Name of the network
        site_filepath: Alternative site info file. Defaults to openghg_defs input.
    Returns:
        list: List of site codes.
    """
    # Load in site data
    site_data = get_site_info(site_filepath=site_filepath)

    network = network.upper()

    matching_sites = []
    for site, details in site_data.items():
        networks = details.keys()
        networks = [n.upper() for n in networks]
        if network in networks:
            matching_sites.append(site)

    return matching_sites
